# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0

import io
import logging
import ntpath

from typing import Dict, Tuple, Optional, List, Generator, Union, Callable

import pefile

from volatility3.framework import interfaces, exceptions
from volatility3.framework import renderers, constants, objects
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows import pdbutil
from volatility3.framework.symbols.windows.extensions import pe
from volatility3.plugins.windows import pslist, modules
from volatility3.framework.constants import windows

vollog = logging.getLogger(__name__)

# keys for specifying wanted names and/or addresses
# used for consistent access between the API and plugins
wanted_names_identifier = "names"
wanted_addresses_identifier = "addresses"

# how wanted modules/symbols are specified, such as:
# {"ntdll.dll" : {wanted_addresses : [42, 43, 43]}}
# {"ntdll.dll" : {wanted_names : ["NtCreateThread"]}}
filter_module_info = Union[Dict[str, List[str]], Dict[str, List[int]]]
filter_modules_type = Dict[str, filter_module_info]

# holds resolved symbols
# {"ntdll.dll": [("Bob", 123), ("Alice", 456)]}
found_symbols_module = List[Tuple[str, int]]
found_symbols_type = Dict[str, found_symbols_module]

# used to hold information about a range (VAD or kernel module)
# (start address, size, file path)
range_type = Tuple[int, int, str]
ranges_type = List[range_type]

# collected_modules are modules and their symbols found when walking vads or kernel modules
# Tuple of (process or kernel layer name, range start, range size)
collected_module_instance = Tuple[str, int, int]
collected_modules_info = List[collected_module_instance]
collected_modules_type = Dict[str, collected_modules_info]

PESymbolFinders = Union[interfaces.context.ModuleInterface, pefile.ExportDirData]


class PESymbolFinder:
    """
    Interface for PE symbol finding classes
    This interface provides a standard way for the calling code to
    lookup symbols by name or address
    """

    cached_str_dict = Dict[str, Optional[str]]

    cached_int_dict = Dict[str, Optional[int]]

    cached_value = Union[int, str, None]
    cached_module_lists = Union[Dict[str, List[str]], Dict[str, List[int]]]
    cached_value_dict = Dict[str, cached_module_lists]

    def __init__(
        self,
        layer_name: str,
        mod_name: str,
        module_start: int,
        symbol_module: PESymbolFinders,
    ):
        self._layer_name = layer_name
        self._mod_name = mod_name
        self._module_start = module_start
        self._symbol_module = symbol_module

        self._address_cache: PESymbolFinder.cached_int_dict = {}
        self._name_cache: PESymbolFinder.cached_str_dict = {}

    def _get_cache_key(self, value: cached_value) -> str:
        """
        Maintain a cache for symbol lookups to avoid re-walking of PDB symbols or export tables
        within the same module for the same address in the same layer

        Args:
            value: The value (address or name) being cached

        Returns:
            str: The constructed cache key that includes the layer and module name
        """
        return f"{self._layer_name}|{self._mod_name}|{value}"

    def get_name_for_address(self, address: int) -> Optional[str]:
        """
        Returns the name for the given address within the particular layer and module

        Args:
            address: the address to resolve within the module

        Returns:
            str: the name of the symbol, if found
        """
        cached_key = self._get_cache_key(address)
        if cached_key not in self._name_cache:
            name = self._do_get_name(address)
            self._name_cache[cached_key] = name

        return self._name_cache[cached_key]

    def get_address_for_name(self, name: str) -> Optional[int]:
        """
        Returns the name for the given address within the particular layer and module

        Args:
            str: the name of the symbol to resolve

        Returns:
            address: the address of the symbol, if found
        """
        cached_key = self._get_cache_key(name)
        if cached_key not in self._address_cache:
            address = self._do_get_address(name)
            self._address_cache[cached_key] = address

        return self._address_cache[cached_key]

    def _do_get_name(self, address: int) -> Optional[str]:
        """
        Returns the name for the given address within the particular layer and module.
        This method must be overwritten by sub classes.

        Args:
            address: the address to resolve within the module

        Returns:
            str: the name of the symbol, if found
        """
        raise NotImplementedError("_do_get_name must be overwritten")

    def _do_get_address(self, name: str) -> Optional[int]:
        """
        Returns the name for the given address within the particular layer and module
        This method must be overwritten by sub classes.

        Args:
            str: the name of the symbol to resolve

        Returns:
            address: the address of the symbol, if found
        """
        raise NotImplementedError("_do_get_address must be overwritten")


class PDBSymbolFinder(PESymbolFinder):
    """
    PESymbolFinder implementation for PDB modules
    """

    def _do_get_address(self, name: str) -> Optional[int]:
        """
        _do_get_address implementation for PDBSymbolFinder

        Args:
            str: the name of the symbol to resolve

        Returns:
            address: the address of the symbol, if found
        """
        try:
            return self._symbol_module.get_absolute_symbol_address(name)
        except exceptions.SymbolError:
            return None

    def _do_get_name(self, address: int) -> Optional[str]:
        """
        _do_get_name  implementation for PDBSymbolFinder

        Args:
            address: the address to resolve within the module

        Returns:
            str: the name of the symbol, if found
        """
        try:
            name = self._symbol_module.get_symbols_by_absolute_location(address)[0]
            return name.split(constants.BANG)[1]
        except (exceptions.SymbolError, IndexError):
            return None


class ExportSymbolFinder(PESymbolFinder):
    """
    PESymbolFinder implementation for PDB modules
    """

    def _get_name(self, export: pefile.ExportData) -> Optional[str]:
        # AttributeError throws on empty or ordinal-only exports
        try:
            return export.name.decode("ascii")
        except AttributeError:
            return None

    def _do_get_name(self, address: int) -> Optional[str]:
        """
        _do_get_name  implementation for ExportSymbolFinder

        Args:
            address: the address to resolve within the module

        Returns:
            str: the name of the symbol, if found
        """
        for export in self._symbol_module:
            if export.address + self._module_start == address:
                return self._get_name(export)

        return None

    def _do_get_address(self, name: str) -> Optional[int]:
        """
        _do_get_address implementation for ExportSymbolFinder
        Args:
            str: the name of the symbol to resolve

        Returns:
            address: the address of the symbol, if found
        """
        for export in self._symbol_module:
            sym_name = self._get_name(export)
            if sym_name and sym_name == name:
                return self._module_start + export.address

        return None


class PESymbols(interfaces.plugins.PluginInterface):
    """Prints symbols in PE files in process and kernel memory"""

    _required_framework_version = (2, 7, 0)

    # 2.0.0 - changed signature of get_kernel_modules, get_all_vads_with_file_paths, addresses_for_process_symbols, get_process_modules
    # 3.0.0 - find_symbols will now throw a ValueError if the provided wanted symbol information does not follow the spec
    _version = (3, 0, 0)

    # used for special handling of the kernel PDB file. See later notes
    os_module_name = "ntoskrnl.exe"

    @classmethod
    def get_requirements(cls) -> List:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="modules", component=modules.Modules, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pdbutil", component=pdbutil.PDBUtility, version=(1, 0, 0)
            ),
            requirements.ChoiceRequirement(
                name="source",
                description="Where to resolve symbols.",
                choices=["kernel", "processes"],
                optional=False,
            ),
            requirements.StringRequirement(
                name="module",
                description='Module in which to resolve symbols. Use "ntoskrnl.exe" to resolve in the base kernel executable.',
                optional=False,
            ),
            requirements.ListRequirement(
                name="symbols",
                element_type=str,
                description="Symbol name to resolve",
                optional=True,
            ),
            requirements.ListRequirement(
                name="addresses",
                element_type=int,
                description="Address of symbol to resolve",
                optional=True,
            ),
        ]

    @classmethod
    def get_pefile_obj(
        cls,
        context: interfaces.context.ContextInterface,
        pe_table_name: str,
        process_layer_name: str,
        base_address: int,
    ) -> Optional[pefile.PE]:
        """
        Attempts to create a pefile object from the bytes of the PE file

        Args:
            pe_table_name: name of the pe types table
            process_layer_name: name of the process layer
            base_address: base address of the module

        Returns:
            the constructed pefile object
        """
        pe_data = io.BytesIO()

        try:
            dos_header = context.object(
                pe_table_name + constants.BANG + "_IMAGE_DOS_HEADER",
                offset=base_address,
                layer_name=process_layer_name,
            )

            for offset, data in dos_header.reconstruct():
                pe_data.seek(offset)
                pe_data.write(data)

            pe_ret = pefile.PE(data=pe_data.getvalue(), fast_load=True)

        except (exceptions.InvalidAddressException, ValueError):
            pe_ret = None

        return pe_ret

    @classmethod
    def range_info_for_address(
        cls, ranges: ranges_type, address: int
    ) -> Optional[range_type]:
        """
        Helper for getting the range information for an address.
        Finds the range holding the `address` parameter

        Args:
            address: the address to find the range for

        Returns:
            Tuple[int, int, str]: The starting address, size, and file path of the range

        """
        for start, size, filepath in ranges:
            if start <= address < start + size:
                return start, size, filepath

        return None

    @classmethod
    def filepath_for_address(cls, ranges: ranges_type, address: int) -> Optional[str]:
        """
        Helper to get the file path for an address

        Args:
            ranges: The set of VADs with mapped files to find the address
            address: The address to find inside of the VADs set

        Returns:
            str: The full path of the file, if found and present
        """
        info = PESymbols.range_info_for_address(ranges, address)
        if info:
            return info[2]

        return None

    @classmethod
    def filename_for_path(cls, filepath: str) -> str:
        """
        Consistent way to get the filename regardless of platform

        Args:
            str: the file path from `filepath_for_address`

        Returns:
            str: the bsae file name of the full path
        """
        return ntpath.basename(filepath).lower()

    @classmethod
    def addresses_for_process_symbols(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        symbols: filter_modules_type,
    ) -> found_symbols_type:
        """
        Used to easily resolve the  addresses of names inside of modules.

        See the usage of this function for system call resolution in  unhooked_system_calls.py
        for an easy to understand example.

        Args:
            symbols: The dictionary of symbols requested by the caller

        Returns:
            found_symbols_type: The dictionary of symbols that were resolved
        """
        collected_modules = PESymbols.get_process_modules(
            context, kernel_module_name, symbols
        )

        found_symbols, missing_symbols = PESymbols.find_symbols(
            context, config_path, symbols, collected_modules
        )

        for mod_name, unresolved_symbols in missing_symbols.items():
            for symbol_key, symbols in unresolved_symbols.items():
                vollog.debug(
                    f"Unable to resolve symbols {symbols} of type {symbol_key} in module {mod_name}"
                )

        return found_symbols

    @classmethod
    def path_and_symbol_for_address(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        collected_modules: collected_modules_type,
        ranges: ranges_type,
        address: int,
    ) -> Tuple[Optional[str], Optional[str]]:
        """
        Method for plugins to determine the file path and symbol name for a given address

        See debugregisters.py for an example of how this function is used along with get_vads_for_process_cache
        for resolving symbols in processes.

        Args:
            collected_modules: return value from `get_kernel_modules` or `get_process_modules`
            ranges: the memory ranges to examine in this layer.
            address: address to resolve to its symbol name
        Returns:
            Tuple[str|renderers.NotApplicableValue|renderers.NotAvailableValue, str|renderers.NotApplicableValue|renderers.NotAvailableValue]
        """
        if not address:
            return None, None

        filepath = PESymbols.filepath_for_address(ranges, address)

        if not filepath:
            return None, None

        filename = PESymbols.filename_for_path(filepath).lower()

        # setup to resolve the address
        filter_module: filter_modules_type = {
            filename: {wanted_addresses_identifier: [address]}
        }

        found_symbols, _missing_symbols = PESymbols.find_symbols(
            context, config_path, filter_module, collected_modules
        )

        if not found_symbols or filename not in found_symbols:
            return filepath, None

        return filepath, found_symbols[filename][0][0]

    @staticmethod
    def _get_exported_symbols(
        context: interfaces.context.ContextInterface,
        pe_table_name: str,
        mod_name: str,
        module_info: collected_module_instance,
    ) -> Optional[ExportSymbolFinder]:
        """
        Attempts to locate symbols based on export analysis

        Args:
            mod_name: lower case name of the module to resolve symbols in
            module_info: (layer_name, module_start, module_size) of the module to examine

        Returns:
            Optional[ExportSymbolFinder]: If the export table can be resolved, then the ExportSymbolFinder
            instance for it
        """

        process_layer_name = module_info[0]
        module_start = module_info[1]

        # we need a valid PE with an export table
        pe_module = PESymbols.get_pefile_obj(
            context, pe_table_name, process_layer_name, module_start
        )
        if not pe_module:
            return None

        pe_module.parse_data_directories(
            directories=[pefile.DIRECTORY_ENTRY["IMAGE_DIRECTORY_ENTRY_EXPORT"]]
        )
        if not hasattr(pe_module, "DIRECTORY_ENTRY_EXPORT"):
            return None

        return ExportSymbolFinder(
            process_layer_name,
            mod_name.lower(),
            module_start,
            pe_module.DIRECTORY_ENTRY_EXPORT.symbols,
        )

    @staticmethod
    def _get_pdb_module(
        context: interfaces.context.ContextInterface,
        config_path: str,
        mod_name: str,
        module_info: collected_module_instance,
    ) -> Optional[PDBSymbolFinder]:
        """
        Attempts to locate symbols based on PDB analysis through each layer where the mod_name module was found

        Args:
            mod_name: lower case name of the module to resolve symbols in
            module_info: (layer_name, module_start, module_size) of the module to examine

        Returns:
            Optional[PDBSymbolFinder]: If the export table can be resolved, then the ExportSymbolFinder
        """

        mod_symbols = None

        layer_name, module_start, module_size = module_info

        # the PDB name of the kernel file is not consistent for an exe, for example,
        # a `ntoskrnl.exe` can have an internal PDB name of any of the ones in the following list
        # The code attempts to find all possible PDBs to ensure the best chance of recovery
        if mod_name == PESymbols.os_module_name:
            pdb_names = [fn + ".pdb" for fn in windows.KERNEL_MODULE_NAMES]

        # for non-kernel files, replace the exe, sys, or dll extension with pdb
        else:
            # in testing we found where some DLLs, such amsi.dll, have its PDB string as Amsi.dll
            # in certain Windows versions
            mod_name = mod_name[:-3] + "pdb"
            first_upper = mod_name[0].upper() + mod_name[1:]
            pdb_names = [mod_name, first_upper]

        # loop through each PDB name (all the kernel names or the dll name as lower() + first char upper case)
        for pdb_name in pdb_names:
            try:
                mod_symbols = pdbutil.PDBUtility.symbol_table_from_pdb(
                    context,
                    interfaces.configuration.path_join(config_path, mod_name),
                    layer_name,
                    pdb_name,
                    module_start,
                    module_size,
                )

                if mod_symbols:
                    break

            # this exception is expected when the PDB can't be found or downloaded
            except exceptions.VolatilityException:
                continue

            # this is not expected - it means pdbconv broke when parsing the PDB
            except TypeError as e:
                vollog.error(
                    f"Unable to parse PDB file for module {pdb_name} -> {e}. Please file a bug on the GitHub issue tracker."
                )

        # cannot do anything without the symbols
        if not mod_symbols:
            return None

        pdb_module = context.module(
            mod_symbols, layer_name=layer_name, offset=module_start
        )

        return PDBSymbolFinder(layer_name, mod_name, module_start, pdb_module)

    @staticmethod
    def _find_symbols_through_pdb(
        context: interfaces.context.ContextInterface,
        config_path: str,
        module_instances: collected_modules_info,
        mod_name: str,
    ) -> Generator[PDBSymbolFinder, None, None]:
        """
        Attempts to resolve the symbols in `mod_name` through PDB analysis

        Args:
            module_instances: the set of layers in which the module was found
            mod_name: name of the module to resolve symbols in
        Returns:
            Generator[PDBSymbolFinder]: a PDBSymbolFinder instance for each layer in which the module was found
        """
        for module_info in module_instances:
            mod_module = PESymbols._get_pdb_module(
                context, config_path, mod_name, module_info
            )
            if mod_module:
                yield mod_module

    @staticmethod
    def _find_symbols_through_exports(
        context: interfaces.context.ContextInterface,
        config_path: str,
        module_instances: collected_modules_info,
        mod_name: str,
    ) -> Generator[ExportSymbolFinder, None, None]:
        """
        Attempts to resolve the symbols in `mod_name` through export analysis

        Args:
            module_instances: the set of layers in which the module was found
            mod_name: name of the module to resolve symbols in
        Returns:
            Generator[ExportSymbolFinder]: an ExportSymbolFinder instance for each layer in which the module was found
        """
        pe_table_name = intermed.IntermediateSymbolTable.create(
            context, config_path, "windows", "pe", class_types=pe.class_types
        )

        # for each process layer and VAD, construct a PE and examine the export table
        for module_info in module_instances:
            exported_symbols = PESymbols._get_exported_symbols(
                context, pe_table_name, mod_name, module_info
            )
            if exported_symbols:
                yield exported_symbols

    @staticmethod
    def _get_symbol_value(
        wanted_symbols: filter_module_info,
        symbol_resolver: PESymbolFinder,
    ) -> Generator[Tuple[str, str, int], None, None]:
        """
        Enumerates the symbols specified as wanted by the calling plugin

        Args:
            wanted_symbols: the set of symbols for a particular module
            symbol_resolver: method in a layer to resolve the symbols

        Returns:
            Tuple[str, str, int]: the symbol identifier (key) of the found symbol in the wanted list, and the name and address of resolved symbol
        """
        if (
            wanted_names_identifier not in wanted_symbols
            and wanted_addresses_identifier not in wanted_symbols
        ):
            vollog.warning(
                "Invalid `wanted_symbols` sent to `find_symbols`. addresses and names keys both missing."
            )
            return

        symbol_keys: List[Tuple[str, Callable]] = [
            (wanted_names_identifier, symbol_resolver.get_address_for_name),
            (wanted_addresses_identifier, symbol_resolver.get_name_for_address),
        ]

        for symbol_key, symbol_getter in symbol_keys:
            # address or name
            if symbol_key in wanted_symbols:
                # walk each wanted address or name

                # build dict in this function for debugging and tracking
                all_wanted = []
                for wanted_value in wanted_symbols[symbol_key]:
                    all_wanted.append(wanted_value)

                for value_index, wanted_value in enumerate(all_wanted):
                    symbol_value = symbol_getter(wanted_value)
                    if symbol_value:
                        # yield out deletion key, deletion index, symbol name, symbol address
                        if symbol_key == wanted_names_identifier:
                            yield symbol_key, wanted_value, symbol_value
                        else:
                            yield symbol_key, symbol_value, wanted_value

                for value in all_wanted:
                    vollog.debug(
                        f"Unable to resolve value {value} using getter {symbol_getter}"
                    )

    @classmethod
    def _validate_wanted_modules(
        cls,
        wanted: PESymbolFinder.cached_module_lists,
    ) -> Optional[PESymbolFinder.cached_module_lists]:
        """
        Validates and makes a copy of the address(es) and/or name(s) wanted from a particular module
        Throws ValueError if invalid values found
        """
        remaining: PESymbolFinder.cached_module_lists = {}

        valid_name_types = [str]
        valid_address_types = [int, objects.Pointer]

        for wanted_type, wanted_symbols in wanted.items():
            if wanted_type not in [
                wanted_names_identifier,
                wanted_addresses_identifier,
            ]:
                raise ValueError(
                    f"The symbol type specified ({wanted_type}) is not valid. Values choices: {wanted_names_identifier}, {wanted_addresses_identifier}"
                )

            remaining[wanted_type] = []

            # symbol_info will be a symbol name or address requested
            for symbol_info in wanted_symbols:
                if wanted_type == wanted_names_identifier and not isinstance(
                    symbol_info, tuple(valid_name_types)
                ):
                    raise ValueError(
                        f"The requested symbol name has a type of {type(symbol_info)} which is not in the allowed set of {valid_name_types}"
                    )

                elif wanted_type == wanted_addresses_identifier and not isinstance(
                    symbol_info, tuple(valid_address_types)
                ):
                    raise ValueError(
                        f"The requested address has a type of {type(symbol_info)} which is not in the allowed set of {valid_address_types}"
                    )

                remaining[wanted_type].append(symbol_info)

        return remaining

    @staticmethod
    def _resolve_symbols_through_methods(
        context: interfaces.context.ContextInterface,
        config_path: str,
        module_instances: collected_modules_info,
        wanted_modules: PESymbolFinder.cached_value_dict,
        mod_name: str,
    ) -> Tuple[found_symbols_module, PESymbolFinder.cached_module_lists]:
        """
        Attempts to resolve every wanted symbol in `mod_name`
        Every layer is enumerated for maximum chance of recovery

        Args:
            module_instances: the set of layers in which the module was found
            wanted_modules: The symbols to resolve tied to their module names
            mod_name: name of the module to resolve symbols in
        Returns:
            Tuple[found_symbols_module, PESymbolFinder.cached_module_lists]: The set of found symbols and the ones that could not be resolved
        """
        symbol_resolving_methods = [
            PESymbols._find_symbols_through_pdb,
            PESymbols._find_symbols_through_exports,
        ]

        found_symbols: found_symbols_module = []

        # the symbols wanted from this module by the caller
        wanted = wanted_modules[mod_name]

        # The ValueError will pass through to the caller
        remaining = PESymbols._validate_wanted_modules(wanted)

        done_processing = False

        for method in symbol_resolving_methods:
            # every layer where this module was found through the given method
            for symbol_resolver in method(
                context, config_path, module_instances, mod_name
            ):
                vollog.debug(f"Have resolver for method {method}")
                for (
                    symbol_key,
                    symbol_name,
                    symbol_address,
                ) in PESymbols._get_symbol_value(remaining, symbol_resolver):
                    found_symbols.append((symbol_name, symbol_address))

                    if symbol_key == wanted_names_identifier:
                        to_remove = symbol_name
                    else:
                        to_remove = symbol_address

                    remaining[symbol_key].remove(to_remove)

                    # everything was resolved, stop this resolver
                    # remove this key from the remaining symbols to resolve
                    if not remaining[symbol_key]:
                        del remaining[symbol_key]
                        done_processing = True
                        break

                if done_processing:
                    break

            # stop all resolving
            if done_processing:
                break

        return found_symbols, remaining

    @classmethod
    def find_symbols(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        wanted_modules: PESymbolFinder.cached_value_dict,
        collected_modules: collected_modules_type,
    ) -> Tuple[found_symbols_type, PESymbolFinder.cached_value_dict]:
        """
        Loops through each method of symbol analysis until each wanted symbol is found
        Returns the resolved symbols as a dictionary that includes the name and runtime address

        `wanted_modules` must be correctly formatted or a ValueError will be thrown

        Args:
            wanted_modules: the dictionary of modules and symbols to resolve. Modified to remove symbols as they are resolved.
            collected_modules: return value from `get_kernel_modules` or `get_process_modules`
        Returns:
            Tuple[found_symbols_type, PESymbolFinder.cached_value_dict]: The set of found symbols but the ones that could not be resolved
        """
        found_symbols: found_symbols_type = {}
        missing_symbols: PESymbolFinder.cached_value_dict = {}

        for mod_name in wanted_modules:
            if mod_name not in collected_modules:
                continue

            module_instances = collected_modules[mod_name]

            # The ValueError from an invalid wanted_modules will pass through to the caller
            # try to resolve the symbols for `mod_name` through each method (PDB and export table currently)
            (
                found_in_module,
                missing_in_module,
            ) = PESymbols._resolve_symbols_through_methods(
                context, config_path, module_instances, wanted_modules, mod_name
            )

            if found_in_module:
                found_symbols[mod_name] = found_in_module

            if missing_in_module:
                missing_symbols[mod_name] = missing_in_module

        return found_symbols, missing_symbols

    @classmethod
    def get_kernel_modules(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_modules: Optional[filter_modules_type],
    ) -> collected_modules_type:
        """
        Walks the kernel module list and finds the session layer, base, and size of each wanted module

        Args:
            filter_modules: The modules to filter the gathering to. If left as None, all kernel modules are gathered.
        Returns:
            collected_modules_type: The collection of modules found with at least one layer present
        """
        found_modules: collected_modules_type = {}

        if filter_modules:
            # create a tuple of module names for use with `endswith`
            filter_modules_check = tuple([key.lower() for key in filter_modules.keys()])
        else:
            filter_modules_check = None

        session_layers = list(
            modules.Modules.get_session_layers(
                context=context, kernel_module_name=kernel_module_name
            )
        )

        # special handling for the kernel
        gather_kernel = (
            filter_modules_check and PESymbols.os_module_name in filter_modules_check
        )

        for index, mod in enumerate(
            modules.Modules.list_modules(context, kernel_module_name)
        ):
            try:
                mod_name = str(mod.BaseDllName.get_string().lower())
            except exceptions.InvalidAddressException:
                continue

            # to analyze, it must either be the kernel or a wanted module
            if not filter_modules_check or (gather_kernel and index == 0):
                mod_name = PESymbols.os_module_name
            elif filter_modules_check and not mod_name.endswith(filter_modules_check):
                continue

            # we won't find symbol information if we can't analyze the module
            session_layer_name = modules.Modules.find_session_layer(
                context, session_layers, mod.DllBase
            )
            if not session_layer_name:
                continue

            if mod_name not in found_modules:
                found_modules[mod_name] = []

            found_modules[mod_name].append(
                (session_layer_name, mod.DllBase, mod.SizeOfImage)
            )

        return found_modules

    @classmethod
    def get_vads_for_process_cache(
        cls,
        vads_cache: Dict[int, ranges_type],
        owner_proc: interfaces.objects.ObjectInterface,
    ) -> Optional[ranges_type]:
        """
        Creates and utilizes a cache of a process' VADs for efficient lookups

        Returns the vad information of the VAD hosting the address, if found

        Args:
            vads_cache: The existing cache of VADs
            owner_proc: The process being inspected
        Returns:
            Optional[ranges_type]: The range holding the address, if found
        """
        if owner_proc.vol.offset in vads_cache:
            vads = vads_cache[owner_proc.vol.offset]
        else:
            vads = PESymbols.get_proc_vads_with_file_paths(owner_proc)
            vads_cache[owner_proc.vol.offset] = vads

        # smear or terminated process
        if len(vads) == 0:
            return None

        return vads

    @classmethod
    def get_proc_vads_with_file_paths(
        cls,
        proc: interfaces.objects.ObjectInterface,
    ) -> ranges_type:
        """
        Returns a list of the process' vads that map a file

        Args:
            proc: The process to gather the VADs for

        Returns:
            ranges_type: The list of VADs for this process that map a file
        """
        vads: ranges_type = []

        try:
            vad_root = proc.get_vad_root()
        except exceptions.InvalidAddressException:
            return vads

        for vad in vad_root.traverse():
            filepath = vad.get_file_name()

            if not isinstance(filepath, str) or filepath.count("\\") == 0:
                continue

            vads.append((vad.get_start(), vad.get_size(), filepath))

        return vads

    @classmethod
    def get_all_vads_with_file_paths(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Generator[
        Tuple[interfaces.objects.ObjectInterface, str, ranges_type],
        None,
        None,
    ]:
        """
        Yields each set of vads for a process that have a file mapped, along with the process itself and its layer

        Args:
            Generator[Tuple[interfaces.objects.ObjectInterface, str, ranges_type]]: Yields tuple of process objects, layers, and VADs mapping files
        """
        procs = pslist.PsList.list_processes(
            context=context,
            kernel_module_name=kernel_module_name,
        )

        for proc in procs:
            try:
                proc_layer_name = proc.add_process_layer()
            except exceptions.InvalidAddressException:
                continue

            vads = cls.get_proc_vads_with_file_paths(proc)

            yield proc, proc_layer_name, vads

    @classmethod
    def get_process_modules(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_modules: Optional[filter_modules_type],
    ) -> collected_modules_type:
        """
        Walks the process list and each process' VAD to determine the base address and size of wanted modules

        Args:
            filter_modules: The modules to filter the gathering to. If left as None, all process modules are gathered.
        Returns:
            collected_modules_type: The collection of modules found with at least one layer present
        """
        proc_modules: collected_modules_type = {}

        if filter_modules:
            # create a tuple of module names for use with `endswith`
            filter_modules_check = tuple([key.lower() for key in filter_modules.keys()])
        else:
            filter_modules_check = None

        for _proc, proc_layer_name, vads in PESymbols.get_all_vads_with_file_paths(
            context, kernel_module_name
        ):
            for vad_start, vad_size, filepath in vads:
                filename = PESymbols.filename_for_path(filepath)

                if filter_modules_check and not filename.endswith(filter_modules_check):
                    continue

                # track each module along with the process layer and range to find it
                if filename not in proc_modules:
                    proc_modules[filename] = []

                proc_modules[filename].append((proc_layer_name, vad_start, vad_size))

        return proc_modules

    def _generator(self) -> Generator[Tuple[int, Tuple[str, str, int]], None, None]:
        if self.config["symbols"]:
            filter_module = {
                self.config["module"].lower(): {
                    wanted_names_identifier: self.config["symbols"]
                }
            }

        elif self.config["addresses"]:
            filter_module = {
                self.config["module"].lower(): {
                    wanted_addresses_identifier: self.config["addresses"]
                }
            }

        else:
            vollog.error("--address or --symbol must be specified")
            return

        if self.config["source"] == "kernel":
            module_resolver = self.get_kernel_modules
        else:
            module_resolver = self.get_process_modules

        collected_modules = module_resolver(
            self.context, self.config["kernel"], filter_module
        )

        found_symbols, _missing_symbols = PESymbols.find_symbols(
            self.context, self.config_path, filter_module, collected_modules
        )

        for module, symbols in found_symbols.items():
            for symbol, address in symbols:
                yield (0, (module, symbol, format_hints.Hex(address)))

    def run(self) -> renderers.TreeGrid:
        return renderers.TreeGrid(
            [
                ("Module", str),
                ("Symbol", str),
                ("Address", format_hints.Hex),
            ],
            self._generator(),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from typing import Tuple


def matches_required(required: Tuple[int, ...], version: Tuple[int, int, int]) -> bool:
    """
    Checks if a version tuple satisfies the required version major and minor constraints.

    Parameters:
        required (Tuple[int, ...]): A tuple containing required major and optionally minor version numbers.
        version (Tuple[int, int, int]): A tuple containing the full version (major, minor, patch).

    Returns:
        bool: True if the version matches the required constraints, False otherwise.
    """
    if len(required) > 0 and version[0] != required[0]:
        return False
    if len(required) > 1 and version[1] < required[1]:
        return False
    return True



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging

from typing import Iterator, Generator, List, Tuple

from volatility3.framework import (
    renderers,
    interfaces,
    constants,
)
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints

vollog = logging.getLogger(__name__)


class KPCRs(interfaces.plugins.PluginInterface):
    """Print KPCR structure for each processor"""

    _required_framework_version = (2, 0, 0)
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
        ]

    @classmethod
    def list_kpcrs(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Generator[Tuple[interfaces.objects.ObjectInterface, int], None, None]:
        """Returns the KPCR structure for each processor

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the kernel module on which to operate

        Returns:
            The _KPCR structure for each processor
        """

        kernel = context.modules[kernel_module_name]
        kernel_layer = context.layers[kernel.layer_name]

        kpcr_type = kernel.get_type("_KPCR")

        reloff = kpcr_type.relative_child_offset("Prcb")

        if kpcr_type.has_member("CurrentPrcb"):
            kpcr_member = "CurrentPrcb"
        else:
            kpcr_member = "Prcb"

        cpu_count_offset = kernel.get_symbol("KeNumberProcessors").address

        cpu_count = kernel.object(
            object_type="unsigned int",
            layer_name=kernel_layer.name,
            offset=cpu_count_offset,
        )

        processor_block = kernel.object(
            object_type="pointer",
            layer_name=kernel_layer.name,
            offset=kernel.get_symbol("KiProcessorBlock").address,
        )

        processor_pointers = utility.array_of_pointers(
            context=context,
            array=processor_block,
            count=cpu_count,
            subtype=kernel.symbol_table_name + constants.BANG + "_KPRCB",
        )

        for pointer in processor_pointers:
            kprcb = pointer.dereference()

            object_address = kprcb.vol.offset - reloff

            if not kernel_layer.is_valid(kprcb.vol.offset):
                continue

            kpcr = kernel.object("_KPCR", offset=object_address, absolute=True)

            yield kpcr, kpcr.member(kpcr_member)

    def _generator(self) -> Iterator[Tuple]:
        for kpcr, current_prcb in self.list_kpcrs(self.context, self.config["kernel"]):
            yield (
                0,
                (
                    format_hints.Hex(kpcr.vol.offset),
                    format_hints.Hex(current_prcb),
                ),
            )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("PRCB Offset", format_hints.Hex),
            ],
            self._generator(),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List, Iterable, Tuple

from volatility3.framework import interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import desktops, windowstations

vollog = logging.getLogger(__name__)


class DeskScan(desktops.Desktops):
    """Scans for the Desktop instances of each Window Station"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.implementation = self.scan_desktops

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="desktops", component=desktops.Desktops, version=(1, 0, 0)
            ),
            requirements.VersionRequirement(
                name="windowstations",
                component=windowstations.WindowStations,
                version=(1, 0, 0),
            ),
        ]

    @classmethod
    def scan_desktops(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
    ) -> Iterable[Tuple[int, str, int, str, str, int]]:
        """
        Yields the information about each desktop and desktop thread needed for analysis

        The tuple yielded includes the:
            Virtual address of the desktop
            The window station name
            The session id
            Desktop name
            Process name
            Process ID (PID)
        """
        kernel = context.modules[kernel_module_name]

        for desktop in windowstations.WindowStations.scan_gui_object(
            context, config_path, kernel_module_name, b"Desk", "tagDESKTOP"
        ):
            desktop_name = desktop.get_name(kernel.symbol_table_name)
            if not desktop_name:
                continue

            winsta = desktop.get_window_station()
            if not winsta:
                continue

            winsta_name, session_id = winsta.get_info(kernel.symbol_table_name)
            if not winsta_name or session_id is None:
                continue

            for _thread, process_name, process_pid in desktop.get_threads():
                yield (
                    format_hints.Hex(desktop.vol.offset),
                    winsta_name,
                    session_id,
                    desktop_name,
                    process_name,
                    process_pid,
                )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Optional, Tuple, Iterator, Generator

from volatility3 import framework
from volatility3.framework import exceptions, constants, interfaces
from volatility3.framework import objects
from volatility3.framework.objects import utility
from volatility3.framework.symbols.windows import extensions
from volatility3.framework.symbols.windows.extensions import pool

vollog = logging.getLogger(__name__)


class GUIExtensions(interfaces.configuration.VersionableInterface):
    _version = (1, 0, 0)
    _required_framework_version = (2, 0, 0)

    framework.require_interface_version(*_required_framework_version)

    class tagWINDOWSTATION(objects.StructType, pool.ExecutiveObject):
        def is_valid(self) -> bool:
            sid = self.get_session_id()
            return sid is not None and 0 <= sid < 256

        def get_session_id(self) -> Optional[int]:
            try:
                return self.dwSessionId
            except exceptions.InvalidAddressException:
                return None

        def traverse(self, max_stations: int = 15):
            """
            Traverses the window stations referenced in the list of stations
            """
            seen = set()

            # include the first window station
            yield self

            while len(seen) < max_stations:
                try:
                    winsta = self.rpwinstaNext.dereference()
                except exceptions.InvalidAddressException:
                    break

                if winsta.vol.offset in seen:
                    break

                yield winsta

                seen.add(winsta.vol.offset)

        def get_info(self, kernel_symbol_table_name) -> Optional[Tuple[str, int]]:
            try:
                name = self.get_name(kernel_symbol_table_name)
                session_id = self.get_session_id()
            except exceptions.InvalidAddressException:
                return None, None

            # attempt to avoid smear
            if session_id is not None and session_id < 256 and name and len(name) > 1:
                return name, session_id

            return None, None

        def desktops(self, symbol_table_name, max_desktops: int = 12):
            seen = set()

            while len(seen) < max_desktops:
                try:
                    desktop = self.rpdeskList.dereference()
                    name = desktop.get_name(symbol_table_name)
                except exceptions.InvalidAddressException:
                    break

                if desktop.vol.offset in seen:
                    break

                yield desktop, name

                seen.add(desktop.vol.offset)

    class tagDESKTOP(objects.StructType, pool.ExecutiveObject):
        def is_valid(self) -> bool:
            """
            Enforce a valid session ID and Window station
            We aren't interested in terminated desktops as there are so many pointers
            going from station -> desktop -> windows, that we would just be processing junk.
            Even if the pointers were still in tact by some miracle, its not that helpful to
            have a floating desktop appear in the output as you can't do much with it.
            """
            sid = self.get_session_id()

            valid_sid = sid is not None and 0 <= sid < 256

            if valid_sid:
                return self.get_window_station() is not None

            return False

        def get_window_station(self) -> Optional["GUIExtensions.tagWINDOWSTATION"]:
            """
            Attempts to return the window station for this desktop
            """
            try:
                return self.rpwinstaParent.dereference()
            except exceptions.InvalidAddressException:
                return None

        def get_session_id(self) -> Optional[int]:
            """
            Attempts to return the session ID for this desktop
            """
            winsta = self.get_window_station()
            if winsta:
                return winsta.get_session_id()

            return None

        def get_threads(
            self,
        ) -> Iterator[Tuple[interfaces.objects.ObjectInterface, str, int]]:
            """
            Returns the threads of each desktop along with owning process information
            """
            symbol_table_name = self.vol.type_name.split(constants.BANG)[0]

            for thread in self.PtiList.to_list(
                symbol_table_name + constants.BANG + "tagTHREADINFO", "PtiLink"
            ):
                try:
                    process_name = utility.array_to_string(
                        thread.ppi.Process.ImageFileName
                    )
                    process_pid = thread.ppi.Process.UniqueProcessId
                except exceptions.InvalidAddressException:
                    continue

                yield thread, process_name, process_pid

        def _do_get_windows(
            self, window, max_windows
        ) -> Generator[Tuple[interfaces.objects.ObjectInterface, str], None, None]:
            """
            Recursively walks and yields the adjacent and child windows
            """
            seen_windows = set()
            seen_children = set()

            if not window.vol.offset:
                return

            yield window, window.get_name()

            seen_windows.add(window)

            # Walk adjacent windows
            while len(seen_windows) < max_windows:
                try:
                    window = window.spwndNext.dereference()
                except exceptions.InvalidAddressException:
                    break

                if not window.vol.offset:
                    break

                if window.vol.offset in seen_windows:
                    break

                yield window, window.get_name()

                seen_windows.add(window)

            # Walk children windows and recursively yield them
            for window in seen_windows:
                child = window

                while len(seen_windows) + len(seen_children) < max_windows:
                    try:
                        child = child.spwndChild
                    except exceptions.InvalidAddressException:
                        break

                    if not child.vol.offset:
                        break

                    if child in seen_children:
                        break
                    seen_children.add(child)

                    yield from self._do_get_windows(child, max_windows)

        def windows(
            self, window, max_windows=10000
        ) -> Generator[Tuple[interfaces.objects.ObjectInterface, str], None, None]:
            """
            Enumerates all windows adjacent to and children of `window`

            Args:
                window: The window to enumerate windows from

            Returns:
                A generator of tuples containing the window and its name
            """
            seen_windows = set()

            for window, window_name in self._do_get_windows(window, max_windows):
                if window.vol.offset in seen_windows:
                    continue

                seen_windows.add(window.vol.offset)

                yield window, window_name

                if len(seen_windows) == max_windows:
                    break

    class tagWND(objects.StructType, pool.ExecutiveObject):
        def is_valid(self) -> bool:
            """
            Enforce a valid sid
            """
            sid = self.get_session_id()

            return sid is not None and 0 <= sid < 256

        def get_name(self) -> Optional[str]:
            """
            directName appeared in later Windows 10 versions and is pointer
            strName is a unicode string directly in the structure
            """
            if self.has_member("directName"):
                try:
                    return utility.pointer_to_string(
                        self.directName, count=256, encoding="utf16"
                    )
                except exceptions.InvalidAddressException:
                    vollog.debug(
                        f"directname for window at {self.vol.offset:#x} in layer {self.vol.layer_name} is invalid"
                    )

            try:
                return self.strName.get_string()
            except exceptions.InvalidAddressException:
                vollog.debug(
                    f"strName for window at {self.vol.offset:#x} in layer {self.vol.layer_name} is invalid"
                )

            return None

        def get_session_id(self) -> Optional[int]:
            """
            Uses its tagDESKTOP pointer to find its session
            """
            desktop = self.get_desktop()
            if desktop:
                return desktop.get_session_id()

            return None

        def get_desktop(self) -> Optional["GUIExtensions.tagDESKTOP"]:
            """
            Attempts to return the host desktop (tagDESKTOP) for this window
            """
            try:
                return self.head.rpdesk.dereference()
            except exceptions.InvalidAddressException:
                vollog.debug(
                    f"Reading the desktop pointer for window {self.vol.offset:#x} caused a page fault"
                )
                return None

        def get_process(self) -> Optional["extensions.EPROCESS"]:
            """
            Attempts to return the host process (_EPROCESS) for this window
            """
            try:
                return self.head.pti.ppi.Process.dereference()
            except exceptions.InvalidAddressException:
                vollog.debug(
                    f"Reading the process pointer for window {self.vol.offset:#x} caused a page fault"
                )
                return None

        def get_window_procedure(self):
            """
            Attempts to return the window procedure for this windows
            """
            try:
                # >= 17134
                if hasattr(self, "subPointer"):
                    return self.subPointer.lpfnWndProc
                else:
                    return self.lpfnWndProc
            except exceptions.InvalidAddressException:
                vollog.debug(
                    f"Invalid window procedure for window {self.vol.offset:#x}"
                )
                return None

    # This is copy/paste from UNICODE_STRING in `symbols/windows/extensions/__init__.py`
    # The versioning of modules would get very ugly if we let different modules share implementations
    # across different data structures
    class LARGE_UNICODE_STRING(objects.StructType):
        """A class for Windows unicode string structures."""

        def get_string(self) -> interfaces.objects.ObjectInterface:
            # We explicitly do *not* catch errors here, we allow an exception to be thrown
            # (otherwise there's no way to determine anything went wrong)
            # It's up to the user of this method to catch exceptions

            # We manually construct an object rather than casting a dereferenced pointer in case
            # the buffer length is 0 and the pointer is a NULL pointer
            return self._context.object(
                self.vol.type_name.split(constants.BANG)[0] + constants.BANG + "string",
                layer_name=self.Buffer.vol.native_layer_name,
                offset=self.Buffer,
                max_length=self.Length,
                errors="replace",
                encoding="utf16",
            )

    class_types = {
        "tagWINDOWSTATION": tagWINDOWSTATION,
        "tagDESKTOP": tagDESKTOP,
        "tagWND": tagWND,
        "_LARGE_UNICODE_STRING": LARGE_UNICODE_STRING,
    }



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import Dict, Iterator, List

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3.framework import deprecation, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.constants import architectures
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols.linux import extensions
from volatility3.framework.symbols.linux.utilities import tainting

vollog = logging.getLogger(__name__)


class Modxview(interfaces.plugins.PluginInterface):
    """Centralize lsmod, check_modules and hidden_modules results to efficiently \
spot modules presence and taints."""

    _version = (1, 0, 0)
    _required_framework_version = (2, 17, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules",
                component=linux_utilities_modules.Modules,
                version=(3, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherer_lsmod",
                component=linux_utilities_modules.ModuleGathererLsmod,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherer_sysfs",
                component=linux_utilities_modules.ModuleGathererSysFs,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherer_scanner",
                component=linux_utilities_modules.ModuleGathererScanner,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux-tainting", component=tainting.Tainting, version=(1, 0, 0)
            ),
            requirements.BooleanRequirement(
                name="plain_taints",
                description="Display the plain taints string for each module.",
                optional=True,
                default=False,
            ),
        ]

    @classmethod
    @deprecation.deprecated_method(
        replacement=linux_utilities_modules.Modules.flatten_run_modules_results,
        replacement_version=(3, 0, 0),
        removal_date="2026-03-25",
    )
    def flatten_run_modules_results(
        cls, run_results: Dict[str, List[extensions.module]], deduplicate: bool = True
    ) -> Iterator[extensions.module]:
        """Flatten a dictionary mapping plugin names and modules list, to a single merged list.
        This is useful to get a generic lookup list of all the detected modules.

        Args:
            run_results: dictionary of plugin names mapping a list of detected modules
            deduplicate: remove duplicate modules, based on their offsets

        Returns:
            Iterator of modules objects
        """
        return linux_utilities_modules.Modules.flatten_run_modules_results(
            run_results, deduplicate
        )

    @classmethod
    @deprecation.deprecated_method(
        replacement=linux_utilities_modules.Modules.run_modules_scanners,
        replacement_version=(3, 0, 0),
        removal_date="2026-03-25",
    )
    def run_modules_scanners(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_name: str,
        run_hidden_modules: bool = True,
    ) -> Dict[str, List[extensions.module]]:
        """Run module scanning plugins and aggregate the results. It is designed
        to not operate any inter-plugin results triage."""
        return linux_utilities_modules.Modules.run_modules_scanners(
            context, kernel_name, run_hidden_modules
        )

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        wanted_gatherers = [
            linux_utilities_modules.ModuleGathererLsmod,
            linux_utilities_modules.ModuleGathererSysFs,
            linux_utilities_modules.ModuleGathererScanner,
        ]

        run_results = linux_utilities_modules.Modules.run_modules_scanners(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            caller_wanted_gatherers=wanted_gatherers,
            flatten=False,
        )

        aggregated_modules = {}
        # We want to be explicit on the plugins results we are interested in
        for gatherer in wanted_gatherers:
            # Iterate over each recovered module
            for mod_info in run_results[gatherer.name]:
                # Use offsets as unique keys, whether a module
                # appears in many plugin runs or not
                if aggregated_modules.get(mod_info.offset, None) is not None:
                    # Append the plugin to the list of originating plugins
                    aggregated_modules[mod_info.offset].append(gatherer.name)
                else:
                    aggregated_modules[mod_info.offset] = [gatherer.name]

        for module_offset, gatherers in aggregated_modules.items():
            module = kernel.object("module", offset=module_offset, absolute=True)

            # Tainting parsing capabilities applied to the module
            if self.config.get("plain_taints"):
                taints = tainting.Tainting.get_taints_as_plain_string(
                    self.context,
                    self.config["kernel"],
                    module.taints,
                    True,
                )
            else:
                taints = ",".join(
                    tainting.Tainting.get_taints_parsed(
                        self.context,
                        self.config["kernel"],
                        module.taints,
                        True,
                    )
                )

            yield (
                0,
                (
                    module.get_name() or renderers.NotAvailableValue(),
                    format_hints.Hex(module_offset),
                    linux_utilities_modules.ModuleGathererLsmod.name in gatherers,
                    linux_utilities_modules.ModuleGathererSysFs.name in gatherers,
                    linux_utilities_modules.ModuleGathererScanner.name in gatherers,
                    taints or renderers.NotAvailableValue(),
                ),
            )

    def run(self):
        columns = [
            ("Name", str),
            ("Address", format_hints.Hex),
            ("In procfs", bool),
            ("In sysfs", bool),
            ("In scan", bool),
            ("Taints", str),
        ]

        return renderers.TreeGrid(
            columns,
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import struct
from typing import Optional, Tuple

from volatility3.framework import exceptions, interfaces, constants
from volatility3.framework.layers import segmented


class LimeFormatException(exceptions.LayerException):
    """Thrown when an error occurs with the underlying Lime file format."""


class LimeLayer(segmented.SegmentedLayer):
    """A Lime format TranslationLayer.

    Lime is generally used to store physical memory images where there
    are large holes in the physical layer
    """

    MAGIC = 0x4C694D45
    VERSION = 1

    # Magic[4], Version[4], Start[8], End[8], Reserved[8]
    # XXX move this to a custom SymbolSpace?
    _header_struct = struct.Struct("<IIQQQ")

    def __init__(
        self, context: interfaces.context.ContextInterface, config_path: str, name: str
    ) -> None:
        super().__init__(context, config_path, name)

        # The base class loads the segments on initialization, but otherwise this must to get the right min/max addresses

    def _load_segments(self) -> None:
        base_layer = self._context.layers[self._base_layer]
        base_maxaddr = base_layer.maximum_address
        maxaddr = 0
        offset = 0
        header_size = self._header_struct.size
        segments = []

        while offset < base_maxaddr:
            start, end = self._check_header(base_layer, offset)

            if start < maxaddr or end < start:
                raise LimeFormatException(
                    self.name,
                    f"Bad start/end 0x{start:x}/0x{end:x} at file offset 0x{offset:x}",
                )

            segment_length = end - start + 1
            segments.append(
                (start, offset + header_size, segment_length, segment_length)
            )
            maxaddr = end
            offset = offset + header_size + segment_length

        if len(segments) == 0:
            raise LimeFormatException(
                self.name, f"No LiME segments defined in {self._base_layer}"
            )

        self._segments = segments

    @classmethod
    def _check_header(
        cls, base_layer: interfaces.layers.DataLayerInterface, offset: int = 0
    ) -> Tuple[int, int]:
        try:
            header_data = base_layer.read(offset, cls._header_struct.size)
        except exceptions.InvalidAddressException:
            raise LimeFormatException(
                base_layer.name,
                f"Offset 0x{offset:0x} does not exist within the base layer",
            )
        (magic, version, start, end, reserved) = cls._header_struct.unpack(header_data)
        if magic != cls.MAGIC:
            raise LimeFormatException(
                base_layer.name, f"Bad magic 0x{magic:x} at file offset 0x{offset:x}"
            )
        if version != cls.VERSION:
            raise LimeFormatException(
                base_layer.name,
                f"Unexpected version {version:d} at file offset 0x{offset:x}",
            )
        return start, end


class LimeStacker(interfaces.automagic.StackerLayerInterface):
    stack_order = 10

    @classmethod
    def stack(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[interfaces.layers.DataLayerInterface]:
        try:
            LimeLayer._check_header(context.layers[layer_name])
        except LimeFormatException:
            return None
        new_name = context.layers.free_layer_name("LimeLayer")
        context.config[interfaces.configuration.path_join(new_name, "base_layer")] = (
            layer_name
        )
        return LimeLayer(context, new_name, new_name)



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
import dataclasses
from typing import List, Iterator

from volatility3.framework import interfaces, renderers, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.renderers import format_hints
from volatility3.framework.constants import architectures
from volatility3.framework.objects import utility
from volatility3.framework.symbols.linux import kallsyms
from volatility3.plugins.linux import pslist

vollog = logging.getLogger(__name__)


@dataclasses.dataclass
class StackEntry:
    position: int
    address: int
    value: int
    name: str = renderers.NotAvailableValue()
    type: str = renderers.NotAvailableValue()
    module: str = renderers.NotAvailableValue()


class PsCallStack(plugins.PluginInterface):
    """Enumerates the call stack of each task"""

    _required_framework_version = (2, 19, 0)

    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="Kallsyms", component=kallsyms.Kallsyms, version=(1, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(4, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="unresolved",
                description="Include unresolved stack values",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    def get_task_callstack(
        cls,
        context: interfaces.context.ContextInterface,
        module_name: str,
        task: interfaces.objects.ObjectInterface,
        kas: kallsyms.Kallsyms = None,
        include_unresolved=False,
    ) -> Iterator[StackEntry]:
        """Retrieves the call stack for a given task

        Args:
            context: The context used to access memory layers and symbols
            module_name: The name of the kernel module on which to operate
            task: The task object whose stack is being retrieved
            kas: Kallsyms instance for symbol resolution. If not provided or None, a new
                instance will be created each time
            include_unresolved: If True, includes stack values that could not be resolved
                to known symbols. Defaults to False.

        Yields:
            StackEntry objects
        """
        task_layer = task.get_address_space_layer()
        if not task_layer:
            return None

        vmlinux = context.modules[module_name]
        vmlinux_layer = context.layers[vmlinux.layer_name]

        if not kas:
            kas = kallsyms.Kallsyms(
                context=context,
                layer_name=vmlinux.layer_name,
                module_name=module_name,
            )

        pointer_size = vmlinux.get_type("pointer").size

        thread_size_order = 2  # Safe since kernel 3.15
        # thread_size_order +=1  # If CONFIG_KASAN is enabled in kernels >= 4.0, default: DISABLED
        # thread_size_order +=1  # If CONFIG_KASAN_EXTRA is enabled in kernels >= 4.19, default: DISABLED
        thread_size = vmlinux_layer.page_size << thread_size_order
        task_base_of_stack = vmlinux_layer.canonicalize(task.stack)
        task_top_of_stack = task_base_of_stack + thread_size

        byte_order = task.files.vol.data_format.byteorder
        rsp_start = task.thread.sp
        if not (task_base_of_stack <= rsp_start < task_top_of_stack):
            raise exceptions.VolatilityException(
                f"Invalid stack pointer {rsp_start:#x} for task {task.pid}"
            )

        current_sp = rsp_start
        idx = 0
        while current_sp < task_top_of_stack:
            try:
                stack_value_bytes = task_layer.read(current_sp, pointer_size)
            except exceptions.InvalidAddressException:
                break
            stack_value = int.from_bytes(stack_value_bytes, byteorder=byte_order)
            if not stack_value:
                idx += 1
                current_sp += pointer_size
                continue
            kassymbol = kas.lookup_address(stack_value)
            sp_address = current_sp & vmlinux_layer.address_mask
            stack_value &= vmlinux_layer.address_mask
            if kassymbol:
                module_name = kassymbol.module_name or renderers.NotAvailableValue()
                yield StackEntry(
                    position=idx,
                    address=sp_address,
                    value=stack_value,
                    name=kassymbol.name,
                    type=kassymbol.type,
                    module=module_name,
                )
            elif include_unresolved:
                yield StackEntry(
                    position=idx,
                    address=sp_address,
                    value=stack_value,
                )

            idx += 1
            current_sp += pointer_size

    def _generator(self):
        module_name = self.config["kernel"]
        vmlinux = self.context.modules[module_name]

        kas = kallsyms.Kallsyms(
            context=self.context,
            layer_name=vmlinux.layer_name,
            module_name=self.config["kernel"],
        )

        include_unresolved = self.config.get("unresolved", False)

        pids = self.config.get("pid", None)
        filter_func = pslist.PsList.create_pid_filter(pids)
        for task in pslist.PsList.list_tasks(
            self.context, vmlinux.name, filter_func=filter_func, include_threads=True
        ):
            task_name = utility.array_to_string(task.comm)

            for stack_entry in self.get_task_callstack(
                context=self.context,
                module_name=vmlinux.name,
                task=task,
                kas=kas,
                include_unresolved=include_unresolved,
            ):
                fields = (
                    task.pid,
                    task_name,
                    stack_entry.position,
                    format_hints.Hex(stack_entry.address),
                    format_hints.Hex(stack_entry.value),
                    stack_entry.name,
                    stack_entry.type,
                    stack_entry.module,
                )
                yield 0, fields

    def run(self):
        return renderers.TreeGrid(
            [
                ("TID", int),
                ("Comm", str),
                ("Position", int),
                ("Address", format_hints.Hex),
                ("Value", format_hints.Hex),
                ("Name", str),
                ("Type", str),
                ("Module", str),
            ],
            self._generator(),
        )



import logging
from typing import List, Union, Tuple

from volatility3.framework import interfaces, renderers, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)


# https://www.ired.team/offensive-security/defense-evasion/masquerading-processes-in-userland-through-_peb
# https://github.com/FuzzySecurity/PowerShell-Suite/blob/master/Masquerade-PEB.ps1
class PebMasquerade(interfaces.plugins.PluginInterface):
    """Detects potential process name spoofing by comparing EPROCESS and PEB data."""

    _version = (1, 0, 0)
    _required_framework_version = (2, 27, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                element_type=int,
                description="Process ID to include (all other processes are excluded)",
                optional=True,
            ),
        ]

    @classmethod
    def get_process_names(
        cls, proc: interfaces.objects.ObjectInterface
    ) -> Tuple[
        Union[str, renderers.NotAvailableValue],
        Union[str, renderers.NotAvailableValue],
        Union[str, renderers.NotAvailableValue],
        Union[str, renderers.NotAvailableValue],
    ]:
        """Extract process names and related information from various sources (EPROCESS and PEB).

        Args:
            proc: The process object

        Returns:
            tuple: (eprocess_imagefilename, eprocess_seaudit_imagefilename, peb_imagefilepath, peb_cmdline)
        """
        eprocess_imagefilename = renderers.NotAvailableValue()
        eprocess_seaudit_imagefilename = renderers.NotAvailableValue()
        peb_imagefilepath = renderers.NotAvailableValue()
        peb_cmdline = renderers.NotAvailableValue()

        try:
            eprocess_imagefilename = utility.array_to_string(proc.ImageFileName)
        except (AttributeError, exceptions.InvalidAddressException):
            vollog.debug(
                "Unable to read EPROCESS.ImageFileName for PID %d", proc.UniqueProcessId
            )
        except Exception as e:
            vollog.warning(
                "Error reading EPROCESS.ImageFileName for PID %d: %s",
                proc.UniqueProcessId,
                str(e),
            )

        try:
            audit = proc.SeAuditProcessCreationInfo.ImageFileName.Name
            audit_string = audit.get_string()
            if audit_string:
                eprocess_seaudit_imagefilename = audit_string
        except exceptions.InvalidAddressException:
            vollog.debug(
                "Unable to read SeAuditProcessCreationInfo.ImageFileName for PID %d",
                proc.UniqueProcessId,
            )
        except AttributeError:
            vollog.debug(
                "SeAuditProcessCreationInfo structure not available for PID %d",
                proc.UniqueProcessId,
            )
        except Exception as e:
            vollog.warning(
                "Error reading SeAuditProcessCreationInfo for PID %d: %s",
                proc.UniqueProcessId,
                str(e),
            )

        try:
            peb = proc.get_peb()
            if peb and peb.ProcessParameters:
                # Get ImagePathName
                try:
                    image_path_str = peb.ProcessParameters.ImagePathName.get_string()
                    if image_path_str:
                        peb_imagefilepath = image_path_str
                except (AttributeError, exceptions.InvalidAddressException):
                    vollog.debug(
                        "Unable to read PEB.ImagePathName for PID %d",
                        proc.UniqueProcessId,
                    )
                except Exception as e:
                    vollog.warning(
                        "Error reading PEB.ImagePathName for PID %d: %s",
                        proc.UniqueProcessId,
                        str(e),
                    )

                try:
                    cmdline_str = peb.ProcessParameters.CommandLine.get_string()
                    if cmdline_str:
                        peb_cmdline = cmdline_str
                except (AttributeError, exceptions.InvalidAddressException):
                    vollog.debug(
                        "Unable to read PEB.ProcessParameters.CommandLine for PID %d",
                        proc.UniqueProcessId,
                    )
                except Exception as e:
                    vollog.warning(
                        "Error reading PEB.ProcessParameters.CommandLine for PID %d: %s",
                        proc.UniqueProcessId,
                        str(e),
                    )
        except (AttributeError, exceptions.InvalidAddressException):
            # Important for cases where PEB does not exist or is inaccessible (e.g SYSTEM process)
            vollog.debug("Unable to access PEB for PID %d", proc.UniqueProcessId)
        except Exception as e:
            vollog.warning(
                "Error accessing PEB for PID %d: %s", proc.UniqueProcessId, str(e)
            )

        return (
            eprocess_imagefilename,
            eprocess_seaudit_imagefilename,
            peb_imagefilepath,
            peb_cmdline,
        )

    def _generator(self, pids, context, kernel_module_name):
        pid_filter = pslist.PsList.create_pid_filter(pids)

        for proc in pslist.PsList.list_processes(
            context=context,
            kernel_module_name=kernel_module_name,
            filter_func=pid_filter,
        ):
            proc_id = proc.UniqueProcessId
            try:
                peb = proc.get_peb()
            except (exceptions.InvalidAddressException, AttributeError):
                vollog.debug(
                    "Unable to access PEB for PID %d, skipping process", proc_id
                )
            peb_imagefilepath_length_check = False
            peb_cmdline_length_check = False
            (
                eprocess_imagefilename,
                eprocess_seaudit_imagefilename,
                peb_imagefilepath,
                peb_cmdline,
            ) = PebMasquerade.get_process_names(proc)

            if isinstance(peb_imagefilepath, str) and peb:
                try:
                    # Length values are of type USHORT
                    peb_imagefilepath_length = (
                        peb.ProcessParameters.ImagePathName.Length // 2
                    )
                    peb_imagefilepath_maxlength = (
                        peb.ProcessParameters.ImagePathName.MaximumLength // 2 - 1
                    )

                    if (peb_imagefilepath_length != len(peb_imagefilepath)) or (
                        peb_imagefilepath_maxlength != len(peb_imagefilepath)
                    ):
                        peb_imagefilepath_length_check = True
                except Exception as e:
                    vollog.warning(
                        "PEB.ImagePathName Length comparison error for PID %d: %s",
                        proc_id,
                        str(e),
                    )

            if isinstance(peb_cmdline, str) and peb:
                try:
                    # Length values are of type USHORT
                    peb_cmdline_length = peb.ProcessParameters.CommandLine.Length // 2
                    peb_cmdline_maxlength = (
                        peb.ProcessParameters.CommandLine.MaximumLength // 2 - 1
                    )

                    if (peb_cmdline_length != len(peb_cmdline)) or (
                        peb_cmdline_maxlength != len(peb_cmdline)
                    ):
                        peb_cmdline_length_check = True
                except Exception as e:
                    vollog.warning(
                        "PEB.CommandLine Length comparison error for PID %d: %s",
                        proc_id,
                        str(e),
                    )
            yield (
                0,
                (
                    proc_id,
                    eprocess_imagefilename,
                    eprocess_seaudit_imagefilename,
                    peb_imagefilepath,
                    peb_cmdline_length_check,
                    peb_imagefilepath_length_check,
                ),
            )

    def run(self):
        pids = self.config.get("pid", None)
        context = self.context
        kernel_module_name = self.config["kernel"]
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("EPROCESS_ImageFileName", str),
                ("EPROCESS_SeAudit_ImageFileName", str),
                ("PEB_ImageFilePath", str),
                ("PEB_ImageFilePath_Spoofed", bool),
                ("PEB_CommandLine_Spoofed", bool),
            ],
            self._generator(pids, context, kernel_module_name),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import dataclasses
import functools
import logging
from typing import Iterator, List, Optional, Tuple

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3.framework import constants, exceptions, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.constants import linux as linux_constants
from volatility3.framework.objects import utility
from volatility3.framework.symbols import linux

vollog = logging.getLogger(__name__)


@dataclasses.dataclass
class KASConfig:
    """Kallsyms configuration class"""

    num_syms_address: int
    names_address: int
    token_table_address: int
    token_index_address: int
    offsets_address: int
    relative_base_address: int
    _stext: int

    # Usually not in VMCOREINFO, these are found during the bootstrap stage.
    # If an ISF is available, they are fetched from there instead.
    markers_address: int = None
    addresses_address: int = None
    _sinittext: int = None
    _einittext: int = None
    _etext: int = None
    _end: int = None
    mod_tree: int = None
    module_addr_min: int = None
    module_addr_max: int = None
    start_ksymtab: int = None
    stop_ksymtab: int = None
    bpf_tree_address: int = None
    seqs_of_names_address: int = None

    num_syms_type_size: int = None
    markers_type_size: int = None
    kernel_symbol_size: int = None

    @classmethod
    def _get_symbol_address(cls, context, layer_name, module_name, symbol_name):
        vmlinux = context.modules[module_name]
        if not vmlinux.has_symbol(symbol_name):
            return None

        layer = context.layers[layer_name]
        address = vmlinux.get_symbol(symbol_name).address
        address += layer.config["kernel_virtual_offset"]
        return address

    @classmethod
    def new_from_isf(cls, context, layer_name, module_name):
        vmlinux = context.modules[module_name]

        # kallsyms_num_syms and kallsyms_markers types were updated from a unsigned long
        # to unsigned int in 4.20 80ffbaa5b1bd98e80e3239a3b8cfda2da433009a
        num_syms_type_size = vmlinux.get_symbol("kallsyms_num_syms").type.size
        kernel_symbol_size = vmlinux.get_type("kernel_symbol").size

        def get_symbol_address(symbol_name):
            return cls._get_symbol_address(
                context, layer_name, module_name, symbol_name
            )

        kas_config = KASConfig(
            num_syms_address=get_symbol_address("kallsyms_num_syms"),
            names_address=get_symbol_address("kallsyms_names"),
            token_table_address=get_symbol_address("kallsyms_token_table"),
            token_index_address=get_symbol_address("kallsyms_token_index"),
            offsets_address=get_symbol_address("kallsyms_offsets"),
            relative_base_address=get_symbol_address("kallsyms_relative_base"),
            markers_address=get_symbol_address("kallsyms_markers"),
            addresses_address=get_symbol_address("kallsyms_addresses"),
            _sinittext=get_symbol_address("_sinittext"),
            _einittext=get_symbol_address("_einittext"),
            _stext=get_symbol_address("_stext"),
            _etext=get_symbol_address("_etext"),
            _end=get_symbol_address("_end"),
            mod_tree=get_symbol_address("mod_tree"),
            module_addr_min=get_symbol_address("module_addr_min"),
            module_addr_max=get_symbol_address("module_addr_max"),
            start_ksymtab=get_symbol_address("__start___ksymtab"),
            stop_ksymtab=get_symbol_address("__stop___ksymtab"),
            bpf_tree_address=get_symbol_address("bpf_tree"),
            seqs_of_names_address=get_symbol_address("kallsyms_seqs_of_names"),
            num_syms_type_size=num_syms_type_size,
            markers_type_size=num_syms_type_size,
            kernel_symbol_size=kernel_symbol_size,
        )
        return kas_config


class _KallsymsIO:
    """Helper to interpret a memory address as a file pointer.

    For internal use within the Kallsyms API; external use is discouraged.
    """

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        base=0,
        endian="little",
    ):
        self._context = context
        self._layer_name = layer_name
        self._base = base
        self._position = base
        self._endian = endian

    def read(self, size: int) -> bytes:
        """Return 'size' bytes from the current position"""
        layer = self._context.layers[self._layer_name]
        buf = layer.read(offset=self._position, length=size)
        self._position += size
        return buf

    def read_str(self, size: int) -> str:
        """Returns 'size' bytes as a string from the current position."""
        return self.read(size).decode()

    def read_int(self, size: int, signed: bool = False) -> int:
        """Returns the integer stored in the current position using 'size' bytes.
        Args:
            size: Number of bytes to use for the int.
            signed: Integer sign.

        Returns:
            The integer stored in the current position.
        """
        return int.from_bytes(
            self.read(size),
            byteorder=self._endian,
            signed=signed,
        )

    def seek(self, offset: int) -> None:
        """Seek the pointer to the given offset, based on the base address.

        Args:
            offset: offset from the base address
        """
        self._position = self._base + offset


@dataclasses.dataclass
class KASSymbolBasic:
    name: str
    type: str


@dataclasses.dataclass
class KASSymbol(KASSymbolBasic):
    address: int
    size: int
    module_name: str
    exported: bool = False
    subsystem: str = None

    def __str__(self):
        return (
            f"name:{self.name}, type:{self.type}, address:{self.address:#x}, "
            f"size:{self.size}, exported:{self.exported}, subsystem:{self.subsystem}"
        )

    def set_exported_from_type(self) -> None:
        """Updates the 'export' member based on the symbol's type.

        This method evaluates the symbol's type and sets the 'export' member
        to indicate whether the object is exported. This code and Linux kernel follows
        the nm symbol type logic.
        """
        # As per the "nm" man page:
        # If lowercase, the symbol is usually local; if uppercase, the symbol is
        # global (external). There are however a few lowercase symbols that are shown
        # for special global symbols ("u", "v" and "w").
        if self.type:
            self.exported = bool(self.type.isupper() or self.type in ("u", "v", "w"))
        else:
            self.exported = None

    @functools.cached_property
    def type_description(self) -> Optional[str]:
        """Returns the interpreted meaning of the symbol type based on the nm tool.

        Returns:
            A string with the type description.
        """
        # If a symbol type exists with the original case, get it
        symbol_type_description = linux_constants.NM_TYPES_DESC.get(self.type, None)
        if symbol_type_description:
            return symbol_type_description

        if self.type:
            # Otherwise, use the lowercase version
            symbol_type_description = linux_constants.NM_TYPES_DESC.get(
                self.type.lower(), None
            )

        return symbol_type_description


@dataclasses.dataclass
class KASFilter:
    name: str
    type: str


class Kallsyms(interfaces.configuration.VersionableInterface):
    """Kallsyms API class"""

    _required_framework_version = (2, 19, 0)
    _version = (1, 0, 0)

    # Internal kernel core constants
    _CORE_SUBSYSTEM_NAME = "core"
    _CORE_MODULE_NAME = "kernel"

    # Internal module constants
    _MODULE_SUBSYSTEM_NAME = "module"

    # Internal FTrace constants
    _FTRACE_SUBSYSTEM_NAME = "ftrace"
    _FTRACE_MODULE_SYM_TYPE = "T"
    _FTRACE_TRAMPOLINE_MODULE_NAME = "__builtin__ftrace"
    _FTRACE_TRAMPOLINE_SYM = "ftrace_trampoline"
    _FTRACE_TRAMPOLINE_SYM_TYPE = "t"

    # Internal BPF constants
    _BPF_SUBSYSTEM_NAME = "bpf"
    _BPF_MODULE_NAME = "bpf"
    _BPF_SYM_TYPE = "t"

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        module_name: str,
        kas_config: KASConfig = None,
        progress_callback: constants.ProgressCallback = None,
    ) -> None:
        """Initialize the Kallsyms API

        Args:
            context: The context used to access memory layers and symbols
            layer_name: The name of layer within the context in which the module exists
            module_name: The name of the kernel module on which to operate
            kas_config: The KAllSyms configuration
            progress_callback: Method that is called periodically during scanning to
                update progress
        """
        super().__init__()

        self._assert_versions()

        self._context = context
        self._layer_name = layer_name
        self._module_name = module_name
        self._kas_config = kas_config
        self._progress_callback = progress_callback
        if progress_callback and not callable(progress_callback):
            raise TypeError("Progress_callback is not callable")

        if not kas_config:
            self._kas_config = KASConfig.new_from_isf(
                context=context,
                layer_name=layer_name,
                module_name=module_name,
            )

        layer = self._context.layers[self._layer_name]
        # FIXME: The layer lacks this information. Could there be a better alternative?
        self._endian = "little" if layer._entry_format[0] == "<" else "big"
        self._long_size = layer.bits_per_register // 8

        self._kallsyms_num_syms = None
        self._kallsyms_relative_base = None

        self._kallsyms_token_index_address = None
        self._kallsyms_offsets_address = None
        self._kallsyms_names_io = _KallsymsIO(
            context=self._context,
            layer_name=self._layer_name,
            base=self._kas_config.names_address,
            endian=self._endian,
        )

        self._kallsyms_token_table_io = _KallsymsIO(
            context=self._context,
            layer_name=self._layer_name,
            base=self._kas_config.token_table_address,
            endian=self._endian,
        )

        self._bootstrap()

    @classmethod
    def _assert_versions(cls) -> None:
        """Verify versions of shared dependencies"""
        linux_utilities_modules_version_required = (3, 0, 0)
        if not requirements.VersionRequirement.matches_required(
            linux_utilities_modules_version_required,
            linux_utilities_modules.Modules.version,
        ):
            raise exceptions.VolatilityException(
                "linux_utilities_modules.Modules version not suitable: "
                f"required {linux_utilities_modules_version_required} found {linux_utilities_modules.Modules.version}",
            )

        return None

    def _read_bytes(self, address: int, size: int) -> Optional[bytes]:
        layer = self._context.layers[self._layer_name]
        try:
            return layer.read(address, size).decode()
        except exceptions.InvalidAddressException:
            return None

    def _read_int(self, address: int, size: int, signed: bool = False) -> Optional[int]:
        layer = self._context.layers[self._layer_name]
        try:
            return int.from_bytes(
                layer.read(address, size),
                byteorder=self._endian,
                signed=signed,
            )
        except exceptions.InvalidAddressException:
            return None

    def _bootstrap(self) -> None:
        layer = self._context.layers[self._layer_name]
        # kallsyms_num_syms and kallsyms_markers[] types were updated from a unsigned long
        # to unsigned int in 4.20 80ffbaa5b1bd98e80e3239a3b8cfda2da433009a
        self._kallsyms_num_syms = self._read_int(
            self._kas_config.num_syms_address,
            self._kas_config.num_syms_type_size,
            signed=False,
        )

        if self._kas_config.relative_base_address:
            # kernels >= 4.6
            self._kallsyms_relative_base = (
                self._read_int(
                    self._kas_config.relative_base_address,
                    self._long_size,
                    signed=False,
                )
                & layer.address_mask
            )

        self._kallsyms_offsets_address = self._kas_config.offsets_address
        self._kallsyms_token_index_address = self._kas_config.token_index_address

        # Preload the kallsyms_token_index array
        short_size = 2
        self._kallsyms_token_index = [
            self._read_int(
                self._kallsyms_token_index_address + index * short_size,
                short_size,
                signed=False,
            )
            for index in range(256)
        ]

    def _get_symbol(
        self,
        offset,
        index,
        filters: List[KASFilter] = None,
    ) -> Optional[Tuple[KASSymbol, int]]:
        kassymbolbasic, compressed_length = self._expand_symbol(offset, filters)
        kassymbol = None
        if kassymbolbasic:
            sym_addr = self._get_symbol_address_by_index(index=index)
            _, sym_size = self._get_symbol_pos(sym_addr)

            kassymbol = KASSymbol(
                name=kassymbolbasic.name,
                type=kassymbolbasic.type,
                address=sym_addr,
                size=sym_size,
                module_name=self._CORE_MODULE_NAME,
                subsystem=self._CORE_SUBSYSTEM_NAME,
            )
            kassymbol.set_exported_from_type()
        return kassymbol, compressed_length

    def get_core_symbols(
        self,
        progress_callback: constants.ProgressCallback = None,
    ) -> Iterator[KASSymbol]:
        """Yield each kernel core symbol

        Args:
            progress_callback: Method that is called periodically during scanning to
                update progress

        Based on kallsyms_on_each_symbol()

        Yields:
            KASSymbol objects
        """
        current_offset = 0
        for sym_idx in range(self._kallsyms_num_syms):
            try:
                kassymbol, compressed_length = self._get_symbol(current_offset, sym_idx)
            except exceptions.InvalidAddressException:
                vollog.debug(
                    f"Unable to reconstruct core symbol at offset {current_offset:#x} and index {sym_idx}"
                )
                continue

            if compressed_length is None:
                vollog.debug(
                    f"Unable to reconstruct compressed_length at offset {current_offset:#x} and index {sym_idx}"
                )
                break

            if kassymbol:
                yield kassymbol

            if progress_callback:
                progress_callback(
                    (sym_idx / self._kallsyms_num_syms) * 100,
                    "Populating Kallsyms core symbols",
                )

            current_offset += compressed_length + 1

    def _expand_symbol(
        self,
        offset: int,
        filters: List[KASFilter] = None,
    ) -> Tuple[KASSymbolBasic, int]:
        """Expand a compressed symbol using its offset in the stream
        Based on kallsyms_expand_symbol()

        Args:
            offset: Symbol offset in the kallsyms arrays.
            filters: List of KASFilter filters

        Returns:
            A tuple with a KASSymbolBasic object and the symbol name's compressed length.
        """
        filters = filters if filters is not None else []
        type_filters = tuple(kassymbolfilter.type for kassymbolfilter in filters)

        self._kallsyms_names_io.seek(offset)
        # The compressed symbol length is in the first byte
        compressed_length = self._kallsyms_names_io.read_int(size=1)
        if compressed_length & 0x80 != 0:
            # kernels >= 6.1 73bbb94466fd3f8b313eeb0b0467314a262dddb3
            # MSB 1 means a 'big' symbol, we need an extra byte
            lower_byte = compressed_length
            upper_byte = self._kallsyms_names_io.read_int(size=1)
            compressed_length = (upper_byte << 7) | (lower_byte & 0x7F)

        abort_decompression = False
        sym_type = None
        sym_name = ""
        for _ in range(compressed_length):
            token_index_index = self._kallsyms_names_io.read_int(size=1)
            token_index = self._kallsyms_token_index[token_index_index]
            self._kallsyms_token_table_io.seek(token_index)
            token = self._kallsyms_token_table_io.read_str(1)
            while token != "\x00":
                if not sym_type:
                    sym_type = token
                    # We got the symbol type, we can abort this immediatelly
                    if type_filters and sym_type not in type_filters:
                        abort_decompression = True
                        break
                else:
                    sym_name += token
                    for kassymbolfilter in filters:
                        if kassymbolfilter.type is not None:
                            if (
                                sym_type == kassymbolfilter.type
                                and kassymbolfilter.name.startswith(sym_name)
                            ):
                                break
                        elif kassymbolfilter.name.startswith(sym_name):
                            break

                    else:
                        if filters:
                            abort_decompression = True

                token = self._kallsyms_token_table_io.read_str(1)

            if abort_decompression:
                break

        kassymbolbasic = (
            KASSymbolBasic(name=sym_name, type=sym_type)
            if not abort_decompression
            else None
        )
        return kassymbolbasic, compressed_length

    def _get_symbol_address_by_index(self, index: int) -> Optional[int]:
        """Return symbol address based on the symbol index in the kallsyms arrays.
        Based on kallsyms_sym_address()

        Args:
            index: Symbol index

        Returns:
            Symbol address
        """
        layer = self._context.layers[self._layer_name]
        if self._kallsyms_offsets_address:
            # kernels >= 4.6 - Addresses are relative to kallsyms_relative_base
            # It assumes: CONFIG_KALLSYMS_BASE_RELATIVE=y and CONFIG_KALLSYMS_ABSOLUTE_PERCPU=y
            signed_int_size = 4
            sym_offset_ptr = self._kallsyms_offsets_address + (index * signed_int_size)
            sym_addr = self._read_int(sym_offset_ptr, signed_int_size, signed=True)
            if sym_addr is None:
                return None

            if sym_addr < 0:
                # Negative offsets are relative to kallsyms_relative_base - 1
                return self._kallsyms_relative_base - 1 - sym_addr

            # Positive offsets are absolute values
            return sym_addr & layer.address_mask
        elif self._kas_config.addresses_address:
            # kernels < 4.6 - Addresses are absolute
            # unsigned long kallsyms_addresses[]
            kallsyms_address = self._read_int(
                self._kas_config.addresses_address + (index * self._long_size),
                self._long_size,
                signed=False,
            )
            if kallsyms_address is None:
                return None

            return kallsyms_address & layer.address_mask
        else:
            raise exceptions.VolatilityException("Unsupported kernel")

    @functools.lru_cache
    def _get_symbol_pos(self, address: int) -> Optional[Tuple[int, int]]:
        """Returns the symbol position in the kallsyms arrays and its size."""
        low = 0
        high = self._kallsyms_num_syms

        while high - low > 1:
            mid = low + (high - low) // 2
            symbol_index = self._get_symbol_address_by_index(mid)
            if symbol_index is None:
                return None, None
            elif symbol_index <= address:
                low = mid
            else:
                high = mid

        # prevent accidental bleed through
        symbol_index = None

        # Search for the first aliased symbol. *Aliased symbols* are symbols with the same address.
        while low:
            symbol_index = self._get_symbol_address_by_index(low - 1)
            if symbol_index is None:
                return None, None

            if symbol_index == self._get_symbol_address_by_index(low):
                low -= 1
            else:
                break

        symbol_start = self._get_symbol_address_by_index(low)
        if symbol_start is None:
            return None, None

        symbol_end = 0

        # Search for next non-aliased symbol.
        for idx in range(low + 1, self._kallsyms_num_syms):
            symbol_index = self._get_symbol_address_by_index(idx)
            if symbol_index is None:
                return None, None

            if symbol_index > symbol_start:
                symbol_end = self._get_symbol_address_by_index(idx)
                break

        # pylint: disable=protected-access
        # If no next symbol is found, we default to using the end of the section
        if not symbol_end:
            if self._is_kernel_inittext(address):
                symbol_end = self._kas_config._einittext
            elif self._kas_config._end is not None:
                # Assume CONFIG_KALLSYMS_ALL=y. Otherwise, symbol_end will be _etext
                symbol_end = self._kas_config._end
            else:
                symbol_end = self._kas_config._etext

        symbol_size = symbol_end - symbol_start

        return low, symbol_size

    @functools.lru_cache
    def _get_symbol_offset(self, index: int) -> int:
        """Find the offset on the compressed stream given the index in the kallsyms array.

        Based on get_symbol_offset

        Returns:
            Offset on the compressed stream
        """

        # Use the nearest marker, placed every 256 positions
        kallsyms_markers_pos_ptr = (
            self._kas_config.markers_address
            + (index >> 8) * self._kas_config.markers_type_size
        )
        kallsyms_markers_pos = self._read_int(
            kallsyms_markers_pos_ptr, self._kas_config.markers_type_size, signed=False
        )
        name_addr = self._kas_config.names_address + kallsyms_markers_pos

        # Scan symbols sequentially until the target. Each symbol uses a
        # [<len>][<len> bytes of data] format, so we skip symbols by adding their length
        # to the pointer value.
        for _ in range(index & 0xFF):
            compressed_length = self._read_int(name_addr, 1)
            if compressed_length & 0x80 != 0:
                # kernels >= 6.1 73bbb94466fd3f8b313eeb0b0467314a262dddb3
                # MSB 1 means a 'big' symbol, we need an extra byte
                lower_byte = compressed_length
                upper_byte = self._kallsyms_names_io.read_int(size=1)
                compressed_length = (upper_byte << 7) | (lower_byte & 0x7F)

            name_addr += compressed_length + 1

        return name_addr - self._kas_config.names_address

    def _is_kernel_inittext(self, addr: int) -> bool:
        # pylint: disable=protected-access
        if not (self._kas_config._sinittext and self._kas_config._einittext):
            # We don't know
            return False

        return self._kas_config._sinittext <= addr < self._kas_config._einittext

    def _is_kernel_text(self, addr: int) -> bool:
        # pylint: disable=protected-access
        return self._kas_config._stext <= addr < self._kas_config._etext

    def _is_core_ksym_addr(self, addr: int) -> bool:
        return self._is_kernel_text(addr) or self._is_kernel_inittext(addr)

    def lookup_address(self, address: int) -> Optional[KASSymbol]:
        """Search for a symbol by its memory address.

        This function scans kernel core, module symbols, BPF symbols, and Ftrace symbols
        to locate the first symbol matching the specified address. Note that multiple
        symbols (aliased symbols) can share the same memory address, so this method
        returns the first match found.

        Based on kallsyms_lookup.

        Args:
            address: The memory address to search for.

        Returns:
            The matching symbol if found, or None if no match is found.
        """
        layer = self._context.layers[self._layer_name]
        address &= layer.address_mask

        kassymbol = self.core_lookup_address(address)
        if not kassymbol:
            kassymbol = self.module_lookup_address(address)

        if not kassymbol:
            kassymbol = self.bpf_lookup_address(address)

        if not kassymbol:
            kassymbol = self.ftrace_lookup_address(address)

        return kassymbol

    def core_lookup_address(self, address: int) -> Optional[KASSymbol]:
        """Search for a symbol by its memory address within the kernel core.

        Based on kallsyms_lookup_buildid.

        Args:
            address: The memory address to search for.

        Returns:
            The matching symbol if found, or None if no match is found.
        """
        layer = self._context.layers[self._layer_name]
        address &= layer.address_mask

        if not self._is_core_ksym_addr(address):
            return None

        pos, sym_size = self._get_symbol_pos(address)
        if pos is None:
            return None
        offset = self._get_symbol_offset(pos)
        sym_address = self._get_symbol_address_by_index(pos)
        kassymbolbasic, _compressed_length = self._expand_symbol(offset)

        if not kassymbolbasic:
            return None

        kas_symbol = KASSymbol(
            name=kassymbolbasic.name,
            type=kassymbolbasic.type,
            address=sym_address,
            size=sym_size,
            module_name=self._CORE_MODULE_NAME,
            subsystem=self._CORE_SUBSYSTEM_NAME,
        )
        kas_symbol.set_exported_from_type()
        return kas_symbol

    def _is_symbol_exported(
        self,
        name: int,
        address: int,
        module: Optional[interfaces.objects.ObjectInterface] = None,
    ) -> bool:
        """Check if the address belongs to an exported symbol.
        If a module object is provided, it searches in that module symbols.
        Otherwise, it searches in the global symbols.

        Bases on is_exported

        Args:
            name: Symbol name
            address: Symbol address
            module: Module object. Defaults to None.

        Returns:
            True if the symbol is exported; otherwise, returns False
        """
        if module:
            if module.num_syms <= 0:
                return False

            start_mod_ksymtab = module.syms
            stop_mod_ksymtab = (
                start_mod_ksymtab
                + module.num_syms * self._kas_config.kernel_symbol_size
            )
            kernel_symbol = self._find_exported_symbol_in_range(
                name, start_mod_ksymtab, stop_mod_ksymtab
            )
        else:
            # Search the not GPL modules
            kernel_symbol = self._find_exported_symbol_in_range(
                name,
                self._kas_config.start_ksymtab,
                self._kas_config.stop_ksymtab,
            )

        if kernel_symbol is not None:
            if hasattr(kernel_symbol, "get_value"):
                return kernel_symbol.get_value() == address
            else:
                return kernel_symbol.vol.offset == address

        return None

    def _elfsym_to_kassymbol(
        self,
        module: interfaces.objects.ObjectInterface,
        elf_sym_obj: interfaces.objects.ObjectInterface,
        elf_sym_index: int,
        subsystem: str = None,
    ) -> Optional[KASSymbol]:
        """Returns a KASSymbol from a ElfSym

        Args:
            module: Module object
            elf_sym_obj: ElfSym object
            elf_sym_index: ElfSym index
            subsystem: Name of the sub-subtem: core, module, bpf, ftrace, etc

        Returns:
            A KASSymbol object
        """
        layer = self._context.layers[self._layer_name]
        sym_name = elf_sym_obj.get_name()
        if not sym_name:
            return None

        # Normalize sym.st_value offset, which is an address pointing to the symbol value
        sym_address = elf_sym_obj.st_value & layer.address_mask
        sym_type = module.get_symbol_type(elf_sym_obj, elf_sym_index)

        kas_symbol = KASSymbol(
            name=sym_name,
            type=sym_type,
            address=sym_address,
            size=elf_sym_obj.st_size,
            module_name=module.get_name(),
            exported=False,
            subsystem=subsystem,
        )
        kas_symbol.set_exported_from_type()
        return kas_symbol

    def _is_module_ksym_address(self, address: int) -> bool:
        return self._modules_address_min <= address <= self._modules_address_max

    def module_lookup_address(
        self,
        address: int,
        module: Optional[interfaces.objects.ObjectInterface] = None,
    ) -> Optional[KASSymbol]:
        """Search for a symbol within kernel modules based on its memory address.
        If a module object is provided, it will only search in that module. Otherwise,
        it will try to first find the module to where the provided address belong to.

        Based on module_address_lookup.

        Args:
            address: The memory address of the symbol to search for
            module [optional]: The module to search within. If not provided, the module
                containing the address will be automatically determined

        Returns:
            The matching KASSymbol if found; otherwise, returns None
        """
        if not self._is_module_ksym_address(address):
            return None

        module = module or self._get_module_by_address(address)
        if not module:
            # This may occur if the kernel lacks the mod_tree implementation.
            for (
                cur_module,
                minimum_address,
                maximum_address,
            ) in self._module_memory_region:
                if minimum_address <= address < maximum_address:
                    module = cur_module
                    break

        if not module:
            # We couldn't find the module
            return None

        kassymbol = self._find_address_in_module_symbols(module, address)
        if kassymbol:
            return kassymbol

        return None

    def _find_address_in_module_symbols(
        self,
        module: interfaces.objects.ObjectInterface,
        address: int,
    ) -> Optional[KASSymbol]:
        """Find the symbol corresponding to a given address within a module.

        Based on find_kallsyms_symbol

        Args:
            module: The module where the address belongs to
            address: The memory address to search for

        Returns:
            The matching KASSymbol if found; otherwise, returns None
        """
        # Before walking all the symbols, ensure the address belongs to this module
        module_boundaries = module.get_module_address_boundaries()
        if not module_boundaries:
            return None

        minimum_address, maximum_address = module_boundaries
        if not (minimum_address <= address < maximum_address):
            return None

        layer = self._context.layers[self._layer_name]
        for elf_sym_idx, elf_sym in enumerate(module.get_symbols()):
            if not elf_sym.get_name():
                continue

            sym_address_start = elf_sym.st_value & layer.address_mask
            sym_address_end = sym_address_start + elf_sym.st_size

            if sym_address_start <= address < sym_address_end:
                return self._elfsym_to_kassymbol(
                    module, elf_sym, elf_sym_idx, subsystem=self._MODULE_SUBSYSTEM_NAME
                )

        return None

    @functools.cached_property
    def _module_memory_region(
        self,
    ) -> List[Tuple[interfaces.objects.ObjectInterface, int, int]]:
        modules_region = []
        for module in linux_utilities_modules.Modules.list_modules(
            self._context, self._module_name
        ):
            minimum_address, maximum_address = module.get_module_address_boundaries()
            module_region = module, minimum_address, maximum_address
            modules_region.append(module_region)

        return modules_region

    @functools.lru_cache
    def _get_modules_memory_boundaries(self) -> Tuple[int, int]:
        """Determine the boundaries of the module allocation area

        Returns:
            A tuple containing the minimum and maximum addresses for the kernel module
            allocation area.
        """

        if self._kas_config.mod_tree:
            # Kernel >= 5.19    58d208de3e8d87dbe196caf0b57cc58c7a3836ca
            mod_tree_address = self._kas_config.mod_tree
            vmlinux = self._context.modules[self._module_name]
            mod_tree = vmlinux.object(
                object_type="mod_tree_root",
                offset=mod_tree_address,
                absolute=True,
            )
            addr_min, addr_max = mod_tree.addr_min, mod_tree.addr_max
        elif self._kas_config.module_addr_min and self._kas_config.module_addr_max:
            # 2.6.27 <= kernel < 5.19   3a642e99babe0617febb6f402e1e063479f489db
            kas_config = self._kas_config
            addr_min, addr_max = kas_config.module_addr_min, kas_config.module_addr_max
        else:
            raise exceptions.VolatilityException(
                "Cannot find the module memory allocation area. Unsupported kernel"
            )

        layer = self._context.layers[self._layer_name]
        return addr_min & layer.address_mask, addr_max & layer.address_mask

    @functools.cached_property
    def _modules_address_min(self):
        address_min, _address_max = self._get_modules_memory_boundaries()
        return address_min

    @functools.cached_property
    def _modules_address_max(self):
        _address_min, address_max = self._get_modules_memory_boundaries()
        return address_max

    def _get_module_by_address(
        self, address: int
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Searches for the module that contains the given memory address within its range.
        It uses a latch tree for optimized address range searching.

            Based on __module_address()

        Args:
            address: The module memory address to search for.

        Returns:
            The matching module if found; otherwise, returns None
        """
        if not self._is_module_ksym_address(address):
            return None

        return self._search_module_by_address(address)

    @functools.lru_cache
    def _get_type_cache(self, name: str) -> Optional[interfaces.objects.Template]:
        vmlinux = self._context.modules[self._module_name]
        try:
            return vmlinux.get_type(name)
        except exceptions.SymbolError:
            return None

    def _mod_tree_comp(
        self, address: int, latch_tree_node: interfaces.objects.ObjectInterface
    ) -> Optional[int]:
        vmlinux = self._context.modules[self._module_name]

        module_memory_mtn = self._get_type_cache("module_memory")
        if not module_memory_mtn:
            vollog.debug(
                "`module_memory` symbol not present in the symbol table. Cannot proceed."
            )
            return None

        module_memory_mtn_offset = module_memory_mtn.relative_child_offset("mtn")

        mod_tree_node_mod = self._get_type_cache("mod_tree_node")
        if not mod_tree_node_mod:
            vollog.debug(
                "`mod_tree_node` symbol not present in the symbol table. Cannot proceed."
            )
            return None

        mod_tree_node_mod_offset = mod_tree_node_mod.relative_child_offset("mod")

        module_memory_offset = (
            latch_tree_node.vol.offset
            + module_memory_mtn_offset
            + mod_tree_node_mod_offset
        )

        module_memory = vmlinux.object(
            object_type="module_memory",
            offset=module_memory_offset,
            absolute=True,
        )
        start = module_memory.base
        end = start + module_memory.size

        if address < start:
            return -1
        elif address >= end:
            return 1
        else:
            return 0

    def _search_module_by_address(
        self, address: int
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Searches for the module that contains the given memory address within its range.
        It uses a latch tree for optimized address range searching.

            Based on mod_find

        Args:
            address: The module memory address to search for

        Returns:
            The matching module if found; otherwise, returns None
        """
        vmlinux = self._context.modules[self._module_name]
        if self._kas_config.mod_tree:
            mod_tree_address = self._kas_config.mod_tree
            mod_tree = vmlinux.object(
                object_type="mod_tree_root",
                offset=mod_tree_address,
                absolute=True,
            )
            latch_tree_root = mod_tree.root
            latch_tree_node = latch_tree_root.find(address, self._mod_tree_comp)
            if latch_tree_node:
                mod_tree_node = linux.LinuxUtilities.container_of(
                    latch_tree_node.vol.offset, "mod_tree_node", "node", vmlinux
                )
                module_ptr = mod_tree_node.mod
                if not module_ptr.is_readable():
                    vollog.warning("Modules latch tree seems corrupt")
                    return None

                return module_ptr.dereference()

        return None

    def _find_exported_symbol_in_range(
        self, name: str, start: int, stop: int
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Find an exported symbol within a specified range of kernel symbols.

        Based on lookup_exported_symbol

        Args:
            name: Symbol name
            start: Start address
            stop: Stop address

        Returns:
            The matching kernel_symbol object if found, or None if no match is found.
        """

        num_elems = (stop - start) // self._kas_config.kernel_symbol_size

        return self._search_kernel_symbol_object_by_name(
            name,
            base=start,
            num_elems=num_elems,
        )

    def _cmp_kernel_symbol_name(
        self,
        name: str,
        kernel_symbol: interfaces.objects.ObjectInterface,
    ) -> int:
        return self._cmp_symbol_name(name, kernel_symbol.get_name())

    def _cmp_symbol_name(
        self,
        name: str,
        other: str,
    ) -> int:
        if name is None or other is None:
            return None
        elif name == other:
            return 0
        elif name < other:
            return -1
        else:
            return 1

    def _search_kernel_symbol_object_by_name(
        self, name: str, base: int, num_elems: int
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Search a kernel_symbol by name using binary search.

        Based on bsearch / __inline_bsearch()

        Args:
            name: Symbol name
            base: Base address
            num_elems: Number of elements

        Returns:
            A kernel_symbol object
        """
        vmlinux = self._context.modules[self._module_name]
        while num_elems > 0:
            pivot = base + (num_elems // 2) * self._kas_config.kernel_symbol_size

            kernel_symbol_pivot = vmlinux.object(
                object_type="kernel_symbol",
                offset=pivot,
                absolute=True,
            )

            result = self._cmp_kernel_symbol_name(name, kernel_symbol_pivot)
            if result == 0:
                return kernel_symbol_pivot
            elif result > 0:
                base = pivot + self._kas_config.kernel_symbol_size
                num_elems -= 1

            num_elems = num_elems // 2

        return None

    def get_modules_symbols(self, name: str = None) -> Iterator[KASSymbol]:
        """Yield each symbol from the kernel modules.
        This function iterates over the symbols of the kernel modules and yields them as
        KASSymbol objects.

        name (optional): If specified, the symbol name used to filter the symbols.

        Yields:
            KASSymbol objects
        """
        layer = self._context.layers[self._layer_name]
        for module in linux_utilities_modules.Modules.list_modules(
            self._context, self._module_name
        ):
            module_name = utility.array_to_string(module.name)
            for elf_sym_idx, elf_sym_obj in enumerate(module.get_symbols()):
                sym_name = elf_sym_obj.get_name()
                if not sym_name:
                    continue

                if name and name != sym_name:
                    continue

                # Normalize sym.st_value offset, which is an address pointing to the symbol value
                sym_address = elf_sym_obj.st_value & layer.address_mask
                sym_size = elf_sym_obj.st_size
                sym_type = module.get_symbol_type(elf_sym_obj, elf_sym_idx)
                is_exported = self._is_symbol_exported(sym_name, sym_address, module)
                sym_type = sym_type.upper() if is_exported else sym_type.lower()

                yield KASSymbol(
                    name=sym_name,
                    type=sym_type,
                    address=sym_address,
                    size=sym_size,
                    exported=is_exported,
                    module_name=module_name,
                    subsystem=self._MODULE_SUBSYSTEM_NAME,
                )

    def _ftrace_mod_get_symbols(self, address: int = None) -> Iterator[KASSymbol]:
        """Yield each symbol from the ftrace modules.
        This function iterates over the symbols of the ftrace modules and yields them as
        KASSymbol objects.

        Based on ftrace_mod_get_kallsym

        Args:
            address (optional): Address to filter symbols by

        Yields:
            KASSymbol objects
        """
        vmlinux = self._context.modules[self._module_name]
        layer = self._context.layers[self._layer_name]
        if not (
            vmlinux.has_type("ftrace_mod_map") and vmlinux.has_type("ftrace_mod_func")
        ):
            # kernel < 4.15 aba4b5c22cbac296f4081a0476d0c55828f135b4
            vollog.info(
                "Unsupported Ftrace kallsyms implementation. Ignore this if it's a kernel < 4.15"
            )
            return None

        symbol_table_name = vmlinux.symbol_table_name
        ftrace_mod_map_symname = f"{symbol_table_name}{constants.BANG}ftrace_mod_map"
        ftrace_mod_func_symname = f"{symbol_table_name}{constants.BANG}ftrace_mod_func"
        ftrace_mod_maps = vmlinux.object_from_symbol("ftrace_mod_maps")
        for mod_map in ftrace_mod_maps.to_list(ftrace_mod_map_symname, "list"):
            for mod_func in mod_map.funcs.to_list(ftrace_mod_func_symname, "list"):
                sym_name = utility.pointer_to_string(
                    mod_func.name, count=linux_constants.KSYM_NAME_LEN
                )
                sym_addr = mod_func.ip & layer.address_mask
                sym_size = mod_func.size
                if address is not None and not (
                    sym_addr <= address < sym_addr + sym_size
                ):
                    continue

                module_name = utility.array_to_string(mod_map.mod.name)
                kas_symbol = KASSymbol(
                    name=sym_name,
                    type=self._FTRACE_MODULE_SYM_TYPE,
                    address=sym_addr,
                    size=sym_size,
                    module_name=module_name,
                    subsystem=self._FTRACE_SUBSYSTEM_NAME,
                )
                kas_symbol.set_exported_from_type()
                yield kas_symbol

    def _ftrace_get_trampoline_symbols(
        self, address: int = None
    ) -> Iterator[KASSymbol]:
        """Yield each symbol from the ftrace trampoline.

        Based on ftrace_get_trampoline_kallsym

        Args:
            address (optional): Address to filter symbols by

        Yields:
            KASSymbol objects
        """
        # See kernel's ftrace_get_trampoline_kallsym()
        vmlinux = self._context.modules[self._module_name]
        if not vmlinux.has_type("ftrace_ops"):
            # kernels < 2.6.27 16444a8a40d4c7b4f6de34af0cae1f76a4f6c901
            return None

        if not vmlinux.has_symbol("ftrace_ops_trampoline_list"):
            # kernels < 5.9 fc0ea795f53c8d7040fa42471f74fe51d78d0834
            return None

        symbol_table_name = vmlinux.symbol_table_name
        ftrace_ops_symname = f"{symbol_table_name}{constants.BANG}ftrace_ops"
        ftrace_ops_trampoline_list = vmlinux.object_from_symbol(
            "ftrace_ops_trampoline_list"
        )

        for ftrace_op in ftrace_ops_trampoline_list.to_list(ftrace_ops_symname, "list"):
            sym_name = self._FTRACE_TRAMPOLINE_SYM
            sym_addr = ftrace_op.trampoline
            sym_size = ftrace_op.trampoline_size

            if address is not None and not (sym_addr <= address < sym_addr + sym_size):
                continue

            kas_symbol = KASSymbol(
                name=sym_name,
                type=self._FTRACE_TRAMPOLINE_SYM_TYPE,
                address=sym_addr,
                size=sym_size,
                module_name=self._FTRACE_TRAMPOLINE_MODULE_NAME,
                subsystem=self._FTRACE_SUBSYSTEM_NAME,
            )
            kas_symbol.set_exported_from_type()
            yield kas_symbol

    def get_ftrace_symbols(self) -> Iterator[KASSymbol]:
        """Yield each kernel ftrace symbol

        Yields:
            KASSymbol objects
        """
        yield from self._ftrace_mod_get_symbols()
        yield from self._ftrace_get_trampoline_symbols()

    def get_bpf_symbols(self) -> Iterator[KASSymbol]:
        """Yield each kernel BPF symbol

        Based on bpf_get_kallsym()

        Yields:
            KASSymbol objects
        """
        vmlinux = self._context.modules[self._module_name]
        if vmlinux.has_type("bpf_ksym"):
            # kernels >= 5.8
            list_type, list_head_member = "bpf_ksym", "lnode"
        elif vmlinux.has_type("bpf_prog_aux"):
            # 3.18 <= kernels < 5.8
            list_type, list_head_member = "bpf_prog_aux", "ksym_lnode"
        else:
            # kernels < 3.18
            vollog.info(
                "Unsupported BPF kallsysms implementation. Don't worry if kernel < 3.18"
            )
            return None

        symbol_table_name = vmlinux.symbol_table_name
        list_type_symname = f"{symbol_table_name}{constants.BANG}{list_type}"

        layer = self._context.layers[self._layer_name]

        # Even when bpf_jit_kallsyms is disabled (/proc/sys/net/core/bpf_jit_kallsyms = 0),
        # this function will still be able to gather the symbols.
        try:
            bpf_kallsyms_list = vmlinux.object_from_symbol("bpf_kallsyms")
        except exceptions.SymbolError:
            vollog.debug(
                "`bpf_kallsyms` symbol not present in the symbol table. Cannot proceed."
            )
            return None

        for elem in bpf_kallsyms_list.to_list(list_type_symname, list_head_member):
            try:
                # See kernel's bpf_get_kallsym()
                if list_type == "bpf_ksym":
                    # kernels >= 5.8
                    bpf_ksym = elem
                    sym_name = utility.array_to_string(bpf_ksym.name)
                    sym_addr = bpf_ksym.start
                    sym_size = bpf_ksym.end - bpf_ksym.start
                else:
                    # list_type == "bpf_prog_aux" 3.18 <= kernels < 5.8
                    bpf_prog_aux = elem
                    bpf_prog = bpf_prog_aux.prog
                    sym_name = bpf_prog.get_name()
                    sym_addr = bpf_prog.bpf_func
                    sym_start, sym_end = bpf_prog.get_address_region()
                    sym_size = sym_end - sym_start
            except exceptions.InvalidAddressException:
                continue

            # The following are also hardcoded in the Linux kernel
            # see kernel's get_ksymbol_bpf(), bpf_get_kallsym() and BPF_SYM_ELF_TYPE
            module_name = self._BPF_MODULE_NAME
            sym_type = self._BPF_SYM_TYPE
            sym_addr &= layer.address_mask

            kas_symbol = KASSymbol(
                name=sym_name,
                type=sym_type,
                address=sym_addr,
                size=sym_size,
                module_name=module_name,
                subsystem=self._BPF_SUBSYSTEM_NAME,
            )
            kas_symbol.set_exported_from_type()
            yield kas_symbol

    def get_all_symbols(self) -> Iterator[KASSymbol]:
        """Enumerates each kallsym symbol

        Yields:
            KASSymbol objects
        """
        yield from self.get_core_symbols()
        yield from self.get_modules_symbols()
        yield from self.get_ftrace_symbols()
        yield from self.get_bpf_symbols()

    def bpf_lookup_address(self, address: int) -> Optional[KASSymbol]:
        """Search for a BPF symbol based on its memory address.

        Based on bpf_address_lookup() and __bpf_address_lookup()

        Args:
            address: The memory address to search for

        Returns:
            The matching KASSymbol if found; otherwise, returns None
        """
        vmlinux = self._context.modules[self._module_name]

        if vmlinux.has_type("bpf_ksym"):
            # kernels >= 5.7 535911c80ad4f5801700e9d827a1985bbff41519
            bpf_ksym = self._find_bpf_ksym(address)
            if not bpf_ksym:
                return None
            symbol_start = bpf_ksym.start
            symbol_end = bpf_ksym.end
            sym_name = utility.array_to_string(bpf_ksym.name)
            sym_size = symbol_end - symbol_start
        elif vmlinux.has_type("latch_tree_root") and vmlinux.get_type(
            "bpf_prog_aux"
        ).has_member("ksym_tnode"):
            # For 4.11 <= kernels < 5.7
            # latch_tree_root was added in kernels 4.2 ade3f510f93a5613b672febe88eff8ea7f1c63b7
            # BPF kallsyms support was added in kernels 4.11 74451e66d516c55e309e8d89a4a1e7596e46aacd
            bpf_prog = self._find_bpf_prog(address)
            if not bpf_prog:
                return None

            symbol_start, symbol_end = bpf_prog.get_addr_region()
            sym_name = bpf_prog.get_name()
            sym_size = symbol_end - symbol_start
        else:
            # kernel < 4.11
            vollog.info(
                "Unsupported BPF kallsyms implementation. Ignore this if it's a kernel < 4.11"
            )
            return None

        layer = self._context.layers[self._layer_name]
        symbol_start &= layer.address_mask

        kas_symbol = KASSymbol(
            name=sym_name,
            type=self._BPF_SYM_TYPE,
            address=symbol_start,
            size=sym_size,
            module_name=self._BPF_MODULE_NAME,
            subsystem=self._BPF_SUBSYSTEM_NAME,
        )
        kas_symbol.set_exported_from_type()
        return kas_symbol

    def _find_bpf_prog(
        self, address: int
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Search for a BPF program based on its address.
        Based on __bpf_address_lookup & bpf_prog_kallsyms_find() for kernels < 5.7

        Args:
            address: The BPF symbol address to search for

        Returns:
            A bpf_prog object if found; otherwise, returns None.
        """
        vmlinux = self._context.modules[self._module_name]
        if not self._kas_config.bpf_tree_address:
            return None

        bpf_latch_tree_root = vmlinux.object(
            object_type="latch_tree_root",
            offset=self._kas_config.bpf_tree_address,
            absolute=True,
        )
        latch_tree_node = bpf_latch_tree_root.find(
            address, self._bpf_tree_comp_bpf_prog_aux
        )

        if not latch_tree_node:
            return None

        bpf_prog_aux = linux.LinuxUtilities.container_of(
            latch_tree_node.vol.offset, "bpf_prog_aux", "ksym_tnode", vmlinux
        )
        bpf_prog = bpf_prog_aux.prog
        return bpf_prog

    def _bpf_tree_comp_bpf_prog_aux(
        self, address: int, latch_tree_node: interfaces.objects.ObjectInterface
    ) -> int:
        """Comparison function used by _find_bpf_prog()
        Based on bpf_tree_comp for kernels < 5.7

        Args:
            address: The memory address to search for
            latch_tree_node: A latch tree node

        Returns:
            0: equal, >0: key is greater, <0: key is less than this bpf_prog
        """
        vmlinux = self._context.modules[self._module_name]
        layer = self._context.layers[self._layer_name]
        bpf_prog_aux = linux.LinuxUtilities.container_of(
            latch_tree_node.vol.offset, "bpf_prog_aux", "ksym_tnode", vmlinux
        )
        bpf_prog = bpf_prog_aux.prog
        bpf_start, bpf_end = bpf_prog.get_address_region()
        bpf_start &= layer.address_mask
        bpf_end &= layer.address_mask

        if address < bpf_start:
            return -1
        elif address > bpf_end:
            # Keep 'key > end' instead of 'key >= end'. This detects return addresses
            # within the program when the final instruction in a stack trace is a call.
            return 1
        else:
            return 0

    def _find_bpf_ksym(
        self, address: int
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Search for the respective bpf_ksym based on a symbol address.
        Based on __bpf_address_lookup & bpf_ksym_find() for kernels >= 5.7

        Args:
            address: The memory address to search for

        Returns:
            A bpf_ksym object if found; otherwise, returns None.
        """
        vmlinux = self._context.modules[self._module_name]
        if not self._kas_config.bpf_tree_address:
            return None

        bpf_latch_tree_root = vmlinux.object(
            object_type="latch_tree_root",
            offset=self._kas_config.bpf_tree_address,
            absolute=True,
        )
        latch_tree_node = bpf_latch_tree_root.find(
            address, self._bpf_tree_comp_bpf_ksym
        )
        if not latch_tree_node:
            return None

        bpf_ksym = linux.LinuxUtilities.container_of(
            latch_tree_node.vol.offset, "bpf_ksym", "tnode", vmlinux
        )
        return bpf_ksym

    def _bpf_tree_comp_bpf_ksym(
        self, address: int, latch_tree_node: interfaces.objects.ObjectInterface
    ) -> int:
        """Comparison function used by _find_bpf_ksym.

        Based on bpf_tree_comp in kernels >= 5.7

        Args:
            address: The memory address to search for
            latch_tree_node: A latch tree node

        Returns:
            0: equal, >0: key is greater, <0: key is less than this bpf_prog
        """
        #
        vmlinux = self._context.modules[self._module_name]
        layer = self._context.layers[self._layer_name]
        bpf_ksym = linux.LinuxUtilities.container_of(
            latch_tree_node.vol.offset, "bpf_ksym", "tnode", vmlinux
        )
        bpf_start = bpf_ksym.start & layer.address_mask
        bpf_end = bpf_ksym.end & layer.address_mask

        if address < bpf_start:
            return -1
        elif address > bpf_end:
            # Keep 'address > bpf_end' instead of 'address >= bpf_end'. This detects return
            # addresses within the program when the final instruction in a stack trace is a call.
            return 1
        else:
            return 0

    def ftrace_lookup_address(self, address: int) -> Optional[KASSymbol]:
        """Search for a ftrace symbol based on its address.

        Based on ftrace_mod_address_lookup()

        Args:
            address: The memory address to search for

        Returns:
            The matching KASSymbol if found, or None if no match is found.
        """

        # Filter by address and return only the first matching result.
        for kassymbol in self._ftrace_mod_get_symbols(address):
            return kassymbol

        for kassymbol in self._ftrace_get_trampoline_symbols(address):
            return kassymbol

        return None

    def _core_lookup_name_slow(self, name) -> Optional[KASSymbol]:
        """Search a core symbol by name

        Based on kallsyms_lookup_name in kernels < 6.2

        Args:
            name: The symbol name to search for.

        Returns:
            A KASSymbol object
        """
        # kernels < 6.2 60443c88f3a89fd303a9e8c0e84895910675c316
        current_offset = 0
        for sym_idx in range(self._kallsyms_num_syms):
            kassymbol, compressed_length = self._get_symbol(current_offset, sym_idx)
            if kassymbol and name == kassymbol.name:
                return kassymbol

            current_offset += compressed_length + 1

        return None

    @functools.cached_property
    def _kallsyms_seqs_of_names(self):
        vmlinux = self._context.modules[self._module_name]
        symbol_table_name = vmlinux.symbol_table_name
        unsigned_char_symname = symbol_table_name + constants.BANG + "unsigned char"
        # See 19bd8981dc2ee35fdc81ab1b0104b607c917d470: 3 bytes per index
        array_size = 3 * self._kallsyms_num_syms
        kallsyms_seqs_of_names = vmlinux.object(
            object_type="array",
            offset=self._kas_config.seqs_of_names_address,
            subtype=vmlinux.get_type(unsigned_char_symname),
            count=array_size,
            absolute=True,
        )
        return kallsyms_seqs_of_names

    def _get_symbol_seq(self, index: int) -> int:
        # See 19bd8981dc2ee35fdc81ab1b0104b607c917d470
        bits = 3
        seq = 0
        for i in range(bits):
            seq = (seq << 8) | self._kallsyms_seqs_of_names[bits * index + i]
        return seq

    def _get_symbol_by_index(self, index) -> Tuple[KASSymbolBasic, int]:
        seq = self._get_symbol_seq(index)
        offset = self._get_symbol_offset(seq)
        kassymbolbasic, _compressed_length = self._expand_symbol(offset)
        return kassymbolbasic

    def _lookup_name_index(self, name: str) -> Optional[int]:
        # based on kallsyms_lookup_names
        high = self._kallsyms_num_syms - 1
        low = 0

        while low <= high:
            mid = (low + high) // 2
            kassymbolbasic = self._get_symbol_by_index(mid)
            if not kassymbolbasic:
                return None

            ret = self._cmp_symbol_name(name, kassymbolbasic.name)
            if ret > 0:
                low = mid + 1
            elif ret < 0:
                high = mid - 1
            else:
                break

        if low > high:
            # Not found
            return None

        low = mid
        while low:
            kassymbolbasic = self._get_symbol_by_index(low - 1)
            if not kassymbolbasic:
                return None
            if self._cmp_symbol_name(name, kassymbolbasic.name) != 0:
                return low
            low -= 1

        return None

    def _core_lookup_name_fast(self, name: str) -> Optional[KASSymbol]:
        """Search a core symbol by name

        Based on kallsyms_lookup_name in kernels >= 6.2

        Args:
            name: The symbol name to search for

        Returns:
            A KASSymbol object
        """
        # kernels >= 6.2 60443c88f3a89fd303a9e8c0e84895910675c316
        index = self._lookup_name_index(name)
        if not index:
            return None

        seq = self._get_symbol_seq(index)
        offset = self._get_symbol_offset(seq)
        kassymbolbasic, _compressed_length = self._expand_symbol(offset)
        sym_address = self._get_symbol_address_by_index(seq)
        _seq, sym_size = self._get_symbol_pos(sym_address)

        kas_symbol = KASSymbol(
            name=kassymbolbasic.name,
            type=kassymbolbasic.type,
            address=sym_address,
            size=sym_size,
            module_name=self._CORE_MODULE_NAME,
            subsystem=self._CORE_SUBSYSTEM_NAME,
        )
        kas_symbol.set_exported_from_type()
        return kas_symbol

    def _kallsyms_lookup_name_modules(self, name: str) -> Optional[KASSymbol]:
        """_summary_

        Based on module_kallsyms_lookup_name

        Args:
            name: The symbol name to search for.

        Returns:
            A KASSymbol object
        """
        for kassymbol in self.get_modules_symbols(name):
            if name == kassymbol.name:
                # First match only
                return kassymbol
        return None

    def lookup_name(self, name: str) -> Optional[KASSymbol]:
        """Search symbols by name.
        WARNING: This function is super slow. The kernel does not index the symbols by
        name, so the it is a linear search.

        Based on kallsyms_lookup_name

        Args:
            name: The symbol name to search for.

        Returns:
            A KASSymbol object
        """
        if self._kas_config.seqs_of_names_address:
            # kernels >= 6.2:
            # 60443c88f3a89fd303a9e8c0e84895910675c316 and 19bd8981dc2ee35fdc81ab1b0104b607c917d470
            kassymbol = self._core_lookup_name_fast(name)
        else:
            # kernels < 6.2
            kassymbol = self._core_lookup_name_slow(name)

        if kassymbol:
            return kassymbol

        return self._kallsyms_lookup_name_modules(name)



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import List, Optional

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3.framework import interfaces, renderers, symbols
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import linux

vollog = logging.getLogger(__name__)


class Check_idt(interfaces.plugins.PluginInterface):
    """Checks if the IDT has been altered"""

    _required_framework_version = (2, 0, 0)

    # 2.0.0 - Add versioning at all, add `get_idt_type`
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules",
                component=linux_utilities_modules.Modules,
                version=(3, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherers",
                component=linux_utilities_modules.ModuleGatherers,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linuxutils", component=linux.LinuxUtilities, version=(2, 0, 0)
            ),
        ]

    @staticmethod
    def get_idt_type(context, vmlinux_name) -> Optional[str]:
        """
        Determines the IDT type for this symbol table or returns None

        The original version ended clauses with an `else` leading to bad fall through
        of returning a type that did not exist in the symbol table.

        Future updates should not leave fall through cases to avoid this repeating.
        """

        vmlinux = context.modules[vmlinux_name]

        is_32bit = not symbols.symbol_table_is_64bit(context, vmlinux.symbol_table_name)

        # These are in a specific order. Only append to the lists going forward
        # or ask Andrew to run tests before merging.
        if is_32bit:
            idt_types = ["gate_struct", "desc_struct", "gate_struct32"]
        else:
            idt_types = ["gate_struct64", "gate_struct", "idt_desc"]

        for idt_type in idt_types:
            if vmlinux.has_type(idt_type):
                return idt_type

        return None

    def _generator(self):
        idt_type = self.get_idt_type(self.context, self.config["kernel"])
        if not idt_type:
            vollog.error(
                "Unable to determine the data structure type for IDT entries. Please file a bug on the GitHub tracker with your kernel version."
            )
            return

        vmlinux = self.context.modules[self.config["kernel"]]

        known_modules = linux_utilities_modules.Modules.run_modules_scanners(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            caller_wanted_gatherers=linux_utilities_modules.ModuleGatherers.all_gatherers_identifier,
        )

        idt_table_size = 256

        kernel_layer = self.context.layers[vmlinux.layer_name]

        address_mask = kernel_layer.address_mask

        # hw handlers + system call
        check_idxs = list(range(20)) + [128]

        addrs = vmlinux.object_from_symbol("idt_table")

        table = vmlinux.object(
            object_type="array",
            offset=addrs.vol.offset,
            subtype=vmlinux.get_type(idt_type),
            count=idt_table_size,
            absolute=True,
        )

        for i in check_idxs:
            ent = table[i]

            if not ent or not kernel_layer.is_valid(ent.vol.offset):
                continue

            if hasattr(ent, "a"):
                idt_addr = (ent.b & 0xFFFF0000) | (ent.a & 0x0000FFFF)
            else:
                low = ent.offset_low
                middle = ent.offset_middle

                # offset_high is for 64bit systems
                if hasattr(ent, "offset_high"):
                    high = ent.offset_high
                else:
                    high = 0

                idt_addr = (high << 32) | (middle << 16) | low

                idt_addr = idt_addr & address_mask

            # 0 means unintialized/unused, not a rootkit
            if idt_addr == 0:
                module_name = renderers.NotAvailableValue()
                symbol_name = renderers.NotAvailableValue()
            else:
                module_info, symbol_name = (
                    linux_utilities_modules.Modules.module_lookup_by_address(
                        self.context, vmlinux.name, known_modules, idt_addr
                    )
                )

                if module_info:
                    module_name = module_info.name
                else:
                    module_name = renderers.NotAvailableValue()

            yield (
                0,
                [
                    format_hints.Hex(i),
                    format_hints.Hex(idt_addr),
                    module_name,
                    symbol_name or renderers.NotAvailableValue(),
                ],
            )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Index", format_hints.Hex),
                ("Address", format_hints.Hex),
                ("Module", str),
                ("Symbol", str),
            ],
            self._generator(),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
import os
from typing import List, Tuple, Iterator, Generator, Dict

from volatility3.framework import interfaces, renderers, symbols, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows import versions
from volatility3.plugins.windows import poolscanner, modules
from volatility3.framework.symbols.windows.extensions import gui

vollog = logging.getLogger(__name__)


class WindowStations(interfaces.plugins.PluginInterface):
    """Scans for top level Windows Stations"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    # These checks must be completed from newest -> oldest OS version.
    _win_version_file_map: List[Tuple[versions.OsDistinguisher, str]] = [
        (versions.is_win10_19577_or_later, "gui-win10-19577-x64"),
        (versions.is_win10_19041_or_later, "gui-win10-19041-x64"),
        (versions.is_win10_18362_or_later, "gui-win10-18362-x64"),
        (versions.is_win10_17763_or_later, "gui-win10-17763-x64"),
        (versions.is_win10_17134_or_later, "gui-win10-17134-x64"),
        (versions.is_win10_16299_or_later, "gui-win10-16299-x64"),
        (versions.is_win10_15063_or_later, "gui-win10-15063-x64"),
        (versions.is_win10_10586_or_later, "gui-win10-10586-x64"),
        (versions.is_windows_8_or_later, "gui-win8-x64"),
        (versions.is_windows_7_sp1, "gui-win7sp1-x64"),
        (versions.is_windows_7_sp0, "gui-win7sp0-x64"),
    ]

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="poolscanner", component=poolscanner.PoolScanner, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="modules", component=modules.Modules, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="GUIExtensions", component=gui.GUIExtensions, version=(1, 0, 0)
            ),
        ]

    @staticmethod
    def create_gui_table(
        context: interfaces.context.ContextInterface,
        symbol_table: str,
        config_path: str,
    ) -> str:
        """Creates a symbol table for windows GUI types

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            symbol_table: The name of an existing symbol table containing the kernel symbols
            config_path: The configuration path within the context of the symbol table to create

        Returns:
            The name of the constructed GUI table
        """

        if not symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=symbol_table
        ):
            raise NotImplementedError(
                "This plugin only supports x64 versions of Windows"
            )

        table_mapping = {"nt_symbols": symbol_table}

        try:
            symbol_filename = next(
                filename
                for version_check, filename in WindowStations._win_version_file_map
                if version_check(context=context, symbol_table=symbol_table)
            )
        except StopIteration:
            raise NotImplementedError("This version of Windows is not supported!")

        vollog.debug(f"Using GUI table {symbol_filename}")

        return intermed.IntermediateSymbolTable.create(
            context=context,
            config_path=config_path,
            sub_path=os.path.join("windows", "gui"),
            filename=symbol_filename,
            class_types=gui.GUIExtensions.class_types,
            table_mapping=table_mapping,
        )

    @classmethod
    def get_session_map(
        cls,
        context: interfaces.context.ContextInterface,
        module_name: str,
        gui_table_name: str,
    ) -> Dict[int, interfaces.context.ModuleInterface]:
        """
        Walks each session layer and returns a dictionary that
        maps session identifiers to a module in the session's layer
        """
        session_map = modules.Modules.get_session_layers_map(context, module_name)

        for session_id, session_layer in session_map.items():
            session_module = context.module(
                gui_table_name,
                layer_name=session_layer,
                offset=context.modules[module_name].offset,
            )
            session_map[session_id] = session_module

        return session_map

    @classmethod
    def scan_gui_object(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        object_tag: bytes,
        object_type: str,
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """
        An API that generically scans for GUI (win32*.sys) objects allocated in the pools (which is nearly all of them)

        This function scans within the kernel space for the tags and then uses `get_session_map` to instantiate objects
        in their correct session address space.

        Args:
            context:
            config_path:
            kernel_module_name:
            object_tag: The 4 byte pool header tag to search for
            object_type: The data structure of the GUI object within the pool
        """

        kernel = context.modules[kernel_module_name]

        gui_table_name = cls.create_gui_table(
            context, kernel.symbol_table_name, config_path
        )

        constraints = poolscanner.PoolScanner.gui_poolscanner_constraints(
            gui_table_name, [object_tag]
        )

        session_map = cls.get_session_map(context, kernel_module_name, gui_table_name)

        for result in poolscanner.PoolScanner.generate_pool_scan_extended(
            context=context,
            kernel_module_name=kernel_module_name,
            object_symbol_table_name=gui_table_name,
            constraints=constraints,
        ):
            _constraint, mem_object, _header = result

            # enforce that objects are in a valid session
            # this prevents smear and also ensures future pointer
            # dereferences are performed in the correct address space (layer)
            try:
                session_id = mem_object.get_session_id()
            except exceptions.InvalidAddressException:
                continue

            if session_id is not None:
                session_module = session_map.get(session_id, None)
                if session_module:
                    # create the object its own address space (per-session)
                    yield session_module.object(
                        object_type=object_type,
                        offset=mem_object.vol.offset,
                        absolute=True,
                    )

    @classmethod
    def scan_window_stations(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
    ) -> Iterator[Tuple["gui.tagWINDOWSTATION", str, int]]:
        """
        Scans for window stations through `scan_gui_object`
        Yields each window station along with its name and session_id
        """

        seen = set()

        kernel = context.modules[kernel_module_name]

        for scanned_winsta in cls.scan_gui_object(
            context, config_path, kernel_module_name, b"Wind", "tagWINDOWSTATION"
        ):
            # walk the list of each station found through scanning
            for winsta in scanned_winsta.traverse():
                if winsta.vol.offset in seen:
                    continue
                seen.add(winsta.vol.offset)

                # stations need to have a name and be in a session
                name, session_id = winsta.get_info(kernel.symbol_table_name)
                if name and session_id is not None:
                    yield winsta, name, session_id

    def _generator(self):
        """
        A wrapper around `scan_window_stations`
        """
        for winsta, name, session_id in self.scan_window_stations(
            self.context, self.config_path, self.config["kernel"]
        ):
            yield (
                0,
                (
                    format_hints.Hex(winsta.vol.offset),
                    name,
                    session_id,
                ),
            )

    # Volatility 2 reported whether the station is interactive or not, but I could not determine if its algorithm
    # is currently valid. I also did not see where the old code paths still checked the same bit mask
    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Name", str),
                ("SessionId", int),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import List, Iterator, Any

from volatility3.framework import exceptions, interfaces
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import mac
from volatility3.plugins.mac import lsmod

vollog = logging.getLogger(__name__)


class Trustedbsd(plugins.PluginInterface):
    """Checks for malicious trustedbsd modules"""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="macutils", component=mac.MacUtilities, version=(1, 3, 0)
            ),
            requirements.VersionRequirement(
                name="lsmod", component=lsmod.Lsmod, version=(2, 0, 0)
            ),
        ]

    def _generator(self, mods: Iterator[Any]):
        kernel = self.context.modules[self.config["kernel"]]

        handlers = mac.MacUtilities.generate_kernel_handler_info(
            self.context, kernel.layer_name, kernel, mods
        )

        policy_list = kernel.object_from_symbol(symbol_name="mac_policy_list").cast(
            "mac_policy_list"
        )

        entries = kernel.object(
            object_type="array",
            offset=policy_list.entries.dereference().vol.offset,
            subtype=kernel.get_type("mac_policy_list_element"),
            absolute=True,
            count=policy_list.staticmax + 1,
        )

        for i, ent in enumerate(entries):
            # I don't know how this can happen, but the kernel makes this check all over the place
            # the policy isn't useful without any ops so a rootkit can't abuse this
            try:
                mpc = ent.mpc.dereference()
                ops = mpc.mpc_ops.dereference()
            except exceptions.InvalidAddressException:
                continue

            try:
                ent_name = utility.pointer_to_string(mpc.mpc_name, 255)
            except exceptions.InvalidAddressException:
                ent_name = "N/A"

            for check in ops.vol.members:
                call_addr = getattr(ops, check)

                if call_addr is None or call_addr == 0:
                    continue

                module_name, symbol_name = mac.MacUtilities.lookup_module_address(
                    self.context, handlers, call_addr, self.config["kernel"]
                )

                yield (
                    0,
                    (
                        check,
                        ent_name,
                        format_hints.Hex(call_addr),
                        module_name,
                        symbol_name,
                    ),
                )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Member", str),
                ("Policy Name", str),
                ("Handler Address", format_hints.Hex),
                ("Handler Module", str),
                ("Handler Symbol", str),
            ],
            self._generator(
                lsmod.Lsmod.list_modules(self.context, self.config["kernel"])
            ),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Iterable, Callable, Tuple

from volatility3.framework import exceptions, renderers, interfaces
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import mac
from volatility3.plugins.mac import pslist

vollog = logging.getLogger(__name__)


class Netstat(plugins.PluginInterface):
    """Lists all network connections for all processes."""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="macutils", component=mac.MacUtilities, version=(1, 0, 0)
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
        ]

    @classmethod
    def list_sockets(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        filter_func: Callable[[int], bool] = lambda _: False,
    ) -> Iterable[
        Tuple[
            interfaces.objects.ObjectInterface,
            interfaces.objects.ObjectInterface,
            interfaces.objects.ObjectInterface,
        ]
    ]:
        """
        Returns the open socket descriptors of a process

        Return values:
            A tuple of 3 elements:
                1) The name of the process that opened the socket
                2) The process ID of the processed that opened the socket
                3) The address of the associated socket structure
        """
        # This is hardcoded, since a change in the default method would change the expected results
        list_tasks = pslist.PsList.get_list_tasks(pslist.PsList.pslist_methods[0])
        for task in list_tasks(context, kernel_module_name, filter_func):
            task_name = utility.array_to_string(task.p_comm)
            pid = task.p_pid

            for filp, _, _ in mac.MacUtilities.files_descriptors_for_process(
                context, context.modules[kernel_module_name].symbol_table_name, task
            ):
                try:
                    ftype = filp.f_fglob.get_fg_type()
                except exceptions.InvalidAddressException:
                    continue

                if ftype != "SOCKET":
                    continue

                try:
                    socket = filp.f_fglob.fg_data.dereference().cast("socket")
                except exceptions.InvalidAddressException:
                    continue

                if not context.layers[task.vol.native_layer_name].is_valid(
                    socket.vol.offset, socket.vol.size
                ):
                    continue

                yield task_name, pid, socket

    def _generator(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        for task_name, pid, socket in self.list_sockets(
            self.context, self.config["kernel"], filter_func=filter_func
        ):
            family = socket.get_family()

            if family == 1:
                try:
                    upcb = socket.so_pcb.dereference().cast("unpcb")
                    path = utility.array_to_string(upcb.unp_addr.sun_path)
                except exceptions.InvalidAddressException:
                    continue

                yield (
                    0,
                    (
                        format_hints.Hex(socket.vol.offset),
                        "UNIX",
                        path,
                        0,
                        "",
                        0,
                        "",
                        f"{task_name}/{pid:d}",
                    ),
                )

            elif family in [2, 30]:
                state = socket.get_state()
                proto = socket.get_protocol_as_string()

                vals = socket.get_converted_connection_info()

                if vals:
                    (lip, lport, rip, rport) = vals

                    yield (
                        0,
                        (
                            format_hints.Hex(socket.vol.offset),
                            proto,
                            lip,
                            lport,
                            rip,
                            rport,
                            state,
                            f"{task_name}/{pid:d}",
                        ),
                    )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Proto", str),
                ("Local IP", str),
                ("Local Port", int),
                ("Remote IP", str),
                ("Remote Port", int),
                ("State", str),
                ("Process", str),
            ],
            self._generator(),
        )
